/*
	This file is part of CrabEmu.

	Copyright (C) 2005 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sms.h"
#include "smsmem.h"
#include "smsvdp.h"
#include "sn76489.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static unsigned char paging_regs[4];
static unsigned char ram[8 * 1024];

static unsigned char *cart_rom;
static unsigned long cart_len;
static unsigned char cart_ram[0x8000];
static unsigned char gg_regs[7];

static unsigned char *rom_page0;
static unsigned char *rom_page1;
static unsigned char *rom_page2;

extern unsigned short sms_pad;
extern sn76489_t psg;
extern int sms_console;

static void reorganize_pages()	{
	rom_page0 = &cart_rom[0x4000 * (paging_regs[1] % (cart_len / 0x4000))];
	rom_page1 = &cart_rom[0x4000 * (paging_regs[2] % (cart_len / 0x4000))];

	if(paging_regs[0] & 0x08)	{
		/* Cartridge RAM enable */
		if(paging_regs[0] & 0x04)	{
			rom_page2 = &cart_ram[0x4000];
		}
		else	{
			rom_page2 = &cart_ram[0];
		}
	}
	else	{
		rom_page2 = &cart_rom[0x4000 * (paging_regs[3] % (cart_len / 0x4000))];
	}
}

void sms_mem_write(unsigned short addr, unsigned char data)	{
	if(addr < 0x8000)	{
		/* Cartridge ROM page 0 and 1, write illegal */
#ifdef DEBUG
		fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
	}
	else if(addr < 0xC000)	{
		if(paging_regs[0] & 0x08)	{
			/* Cartridge RAM page 0 or 1 */
			rom_page2[addr & 0x3FFF] = data;
		}
		else	{
			/* Cartridge ROM page 2, write illegal */
#ifdef DEBUG
			fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
		}
	}
	else	{
		/* On-board RAM */
		ram[addr & 0x1FFF] = data;
		if(addr > 0xFFFB)	{
			/* Paging registers */
			paging_regs[addr - 0xFFFC] = data;
			reorganize_pages();
		}
	}
}

void sms_port_write(unsigned short port, unsigned char data)	{
	if(port < 0x07 && sms_console == CONSOLE_GG)	{
		gg_regs[port] = data;
	}
	else if(port < 0x40)	{
		if(port & 0x01)	{
			/* I/O Control register */
		}
		else	{
			/* Memory Control register */
		}
	}
	else if(port < 0x80)	{
		/* SN76489 PSG */
		sn76489_write(&psg, data);
	}
	else if(port < 0xC0)	{
		if(port & 0x01)	{
			/* VDP Control port */
			sms_vdp_ctl_write(data);
		}
		else	{
			/* VDP Data port */
			sms_vdp_data_write(data);
		}
	}
	else if(port < 0x100)	{
		/* No Effect */
	}
	else	{
		fprintf(stderr, "sms_port_write: Write to invalid port: %04x\n", port);
	}
}

unsigned char sms_mem_read(unsigned short addr)	{
	if(addr < 0x0400)	{
		/* 1k of ROM bank 0 - NEVER paged out! */
		return cart_rom[addr];
	}
	else if(addr < 0x4000)	{
		/* Cartridge ROM page 0 */
		return rom_page0[addr];
	}
	else if(addr < 0x8000)	{
		/* Cartridge ROM page 1 */
		return rom_page1[addr & 0x3FFF];
	}
	else if(addr < 0xC000)	{
		/* Cartridge ROM page 2, or Cartridge RAM page 0 or 1 */
		return rom_page2[addr & 0x3FFF];
	}
	else	{
		/* On-board RAM */
		return ram[addr & 0x1FFF];
	}
}

unsigned char sms_port_read(unsigned short port)	{
	if(port < 0x07 && sms_console == CONSOLE_GG)	{
		return gg_regs[port];
	}
	else if(port < 0x40)	{
		return 0xFF;
	}
	else if(port < 0x80)	{
		if(port & 0x01)	{
			return sms_vdp_hcnt_read();
		}
		else	{
			return sms_vdp_vcnt_read();
		}
	}
	else if(port < 0xC0)	{
		if(port & 0x01)	{
			return sms_vdp_status_read();
		}
		else	{
			return sms_vdp_data_read();
		}
	}
	else if(port < 0x100)	{
		if(port & 0x01)	{
			/* I/O port B/misc register */
			return (sms_pad >> 8) & 0xFF;
		}
		else	{
			/* I/O port A/B register */
			return (sms_pad & 0xFF);
		}
	}
	else	{
		fprintf(stderr, "sms_read_port: Attempt to read invalid port: %04x\n", port);
		return 0;
	}
}

int sms_mem_load_rom(const char *fn)	{
	FILE *fp;
	int size;
	char *ext;

	fp = fopen(fn, "rb");
	if(!fp)	{
		fprintf(stderr, "sms_mem_load_rom: Could not open ROM: %s!\n", fn);
		return -1;
	}

	/* First, determine if its a SMS or a GameGear ROM */
	ext = strrchr(fn, '.');
	if(!strcmp(ext, ".gg"))	{
		/* GameGear rom found */
		sms_console = CONSOLE_GG;
	}
	else	{
		/* Master System rom found */
		sms_console = CONSOLE_SMS;
	}

	/* Determine the size of the file */
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	
	/* Determine if header/footer info is present */
	if((size % 0x4000) == 512)	{
		/* SMD header present, skip it */
		fseek(fp, 512, SEEK_SET);
		size -= 512;
	}
	else if((size % 0x4000) == 64)	{
		/* Footer present, truncate it */
		size -= 64;
	}
	
	cart_rom = (unsigned char *)malloc(size);
	fread(cart_rom, size, 1, fp);
	fclose(fp);

	cart_len = size;
	
	reorganize_pages();
	
	return 0;
}

int sms_mem_init()	{
	cart_rom = NULL;
	rom_page0 = NULL;
	rom_page1 = NULL;
	rom_page2 = NULL;

	paging_regs[0] = 0;
	paging_regs[1] = 0;
	paging_regs[2] = 1;
	paging_regs[3] = 2;

	gg_regs[0] = 0xC0;
	gg_regs[1] = 0x7F;
	gg_regs[2] = 0xFF;
	gg_regs[3] = 0x00;
	gg_regs[4] = 0xFF;
	gg_regs[5] = 0x00;
	gg_regs[6] = 0xFF;

	return 0;
}

int sms_mem_shutdown()	{
	return 0;
}
