/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <string.h>
#include "sms.h"
#include "smsvdp.h"
#include "smsmem.h"
#include "smsz80.h"
#include "sn76489.h"
#include "sound.h"

unsigned short sms_pad = 0xFFFF;
int sms_psg_enabled = 1;
int sms_initialized = 0;
sn76489_t psg;
int sms_console = CONSOLE_SMS;
int sms_region = SMS_REGION_EXPORT;

static unsigned long psg_samples[313];

static const float NTSC_Z80_CLOCK = 3579545.0f;
static const int PSG_DIVISOR = 16;
static const int NTSC_FPS = 60;
static const int NTSC_LINES_PER_FRAME = 262;
static const float NTSC_CLOCKS_PER_SAMPLE = 5.073051303855f;

static const float PAL_Z80_CLOCK = 3546893.0f;
static const int PAL_FPS = 50;
static const int PAL_LINES_PER_FRAME = 313;
static const float PAL_CLOCKS_PER_SAMPLE = 5.02677579365f;

int sms_init(int system_type)   {
    int i;
    float tmp;

    if(system_type == SMS_VIDEO_NTSC)   {
        tmp = NTSC_Z80_CLOCK / PSG_DIVISOR / NTSC_FPS / NTSC_LINES_PER_FRAME /
              NTSC_CLOCKS_PER_SAMPLE;

        for(i = 0; i < NTSC_LINES_PER_FRAME; ++i)   {
            psg_samples[i] = (unsigned long) (tmp * (i + 1)) -
                             (unsigned long) (tmp * i);
        }

        /* We end up generating 734 samples per frame @ 44100 Hz, 60fps, but we
           need 735. */
        psg_samples[261] += 1;

        sms_region |= SMS_VIDEO_NTSC;

        sn76489_init(&psg, NTSC_Z80_CLOCK, 44100.0f, 0x8000, 0x0009);
    }
    else    {
        tmp = PAL_Z80_CLOCK / PSG_DIVISOR / PAL_FPS / PAL_LINES_PER_FRAME /
             PAL_CLOCKS_PER_SAMPLE;

        for(i = 0; i < PAL_LINES_PER_FRAME; ++i)    {
            psg_samples[i] = (unsigned long) (tmp * (i + 1)) -
                             (unsigned long) (tmp * i);            
        }

        /* We need 882 samples per frame @ 44100 Hz, 50fps. */
        sms_region |= SMS_VIDEO_PAL;

        sn76489_init(&psg, PAL_Z80_CLOCK, 44100.0f, 0x8000, 0x0009);
    }

    sms_mem_init();
    sms_vdp_init(system_type);
    sms_z80_init();

    sound_init();

    sms_initialized = 1;

    return 0;
}

int sms_reset() {
    if(sms_initialized == 0)
        return 0;

    sms_mem_shutdown();
    sms_mem_init();

    sms_z80_reset();

    return 0;
}

void sms_soft_reset()   {
    if(sms_initialized == 0)
        return;

    sound_reset_buffer();

    sms_mem_reset();
    sms_z80_reset();
}

int sms_shutdown()  {
    sms_mem_shutdown();
    sms_vdp_shutdown();
    sms_z80_shutdown();
    sound_shutdown();

    return 0;
}

void sms_frame()    {
    signed short buf[882 << 1];
    int line, total_lines;

    if(sms_region & SMS_VIDEO_NTSC)
        total_lines = NTSC_LINES_PER_FRAME;
    else
        total_lines = PAL_LINES_PER_FRAME;

    for(line = 0; line < total_lines; ++line)   {
        sms_z80_run(227);
        sms_vdp_execute(line);

        if(sms_psg_enabled) {
            sn76489_execute_samples(&psg, buf, psg_samples[line]);
            sound_update_buffer(buf, psg_samples[line] << 2);
        }
    }
}

void sms_button_pressed(int button) {
    if(button == GG_START)  {
        if(sms_console == CONSOLE_GG)   {
            uint8 byte = sms_port_read(0);
            byte &= ~0x80;
            sms_port_write(0, byte);
        }
    }
    else    {
        sms_pad &= ~button;
    }
}

void sms_button_released(int button)    {
    if(button == GG_START)  {
        if(sms_console == CONSOLE_GG)   {
            uint8 byte = sms_port_read(0);
            byte |= 0x80;
            sms_port_write(0, byte);
        }
    }
    else    {
        sms_pad |= button;
    }
}

void sms_psg_write_context(FILE *fp) {
    unsigned char byte;
    int i;
    unsigned short counter;

    fwrite(psg.volume, 4, 1, fp);

    for(i = 0; i < 3; ++i)  {
        byte = psg.tone[i] & 0xFF;
        fwrite(&byte, 1, 1, fp);
        byte = (psg.tone[i] >> 8) & 0xFF;
        fwrite(&byte, 1, 1, fp);
    }

    fwrite(&psg.noise, 1, 1, fp);
    fwrite(psg.tone_state, 4, 1, fp);
    fwrite(&psg.latched_reg, 1, 1, fp);

    for(i = 0; i < 4; ++i)  {
        counter = (unsigned short) psg.counter[i];
        byte = counter & 0xFF;
        fwrite(&byte, 1, 1, fp);
        byte = (counter >> 8) & 0xFF;
        fwrite(&byte, 1, 1, fp);
    }
}

void sms_psg_read_context(FILE *fp) {
    unsigned char byte[2];
    int i;
    unsigned short counter;

    fread(psg.volume, 4, 1, fp);

    for(i = 0; i < 3; ++i)  {
        fread(&byte, 2, 1, fp);
        psg.tone[i] = byte[0] | (byte[1] << 8);
    }

    fread(&psg.noise, 1, 1, fp);
    fread(psg.tone_state, 4, 1, fp);
    fread(&psg.latched_reg, 1, 1, fp);

    for(i = 0; i < 4; ++i)  {
        fread(&byte, 2, 1, fp);
        counter = byte[0] | (byte[1] << 8);
        psg.counter[i] = counter;
    }
}

int sms_save_state(const char *filename)   {
    FILE *fp;
    unsigned char byte;

    if(sms_initialized == 0)    {
        /* This shouldn't happen.... */
        return -1;
    }

    fp = fopen(filename, "wb");
    if(!fp)
        return -1;

    fprintf(fp, "CrabEmu Save State");

    /* Write save state version */
    byte = 0x00;
    fwrite(&byte, 1, 1, fp);

    byte = 0x01;
    fwrite(&byte, 1, 1, fp);

    /* Write out the current Z80 context */
    sms_z80_write_context(fp);

    /* Next, write the current VDP state */
    sms_vdp_write_context(fp);

    /* Now, write the current PSG state */
    sms_psg_write_context(fp);

    /* Finally, write the current memory contents to the file */
    sms_mem_write_context(fp);

    fclose(fp);

    return 0;
}

int sms_load_state(const char *filename)   {
    char str[19];
    FILE *fp;
    char byte;

    if(sms_initialized == 0)    {
        /* This shouldn't happen.... */
        return -1;
    }

    fp = fopen(filename, "rb");
    if(!fp)
        return -1;

    fread(str, 18, 1, fp);
    str[18] = 0;
    if(strcmp("CrabEmu Save State", str))   {
        fclose(fp);
        return -2;
    }

    /* Read save state version */
    fread(&byte, 1, 1, fp);
    if(byte != 0x00)    {
        fclose(fp);
        return -2;
    }

    fread(&byte, 1, 1, fp);
    if(byte != 0x01)    {
        fclose(fp);
        return -2;
    }

    /* Read in the current Z80 context */
    sms_z80_read_context(fp);

    /* Next, read the current VDP state */
    sms_vdp_read_context(fp);

    /* Now, read the current PSG state */
    sms_psg_read_context(fp);

    /* Finally, read the current memory contents from the file */
    sms_mem_read_context(fp);

    fclose(fp);

    sound_reset_buffer();

    return 0;
}
