/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "EmuController.h"
#include "EmuOGLView.h"
#include "sms.h"
#include "smsmem.h"
#include "sn76489.h"
#include "sound.h"

/* From main.m */
extern NSLock *memoryLock;
extern EmuOGLView *theview;
extern void emu_loop();
extern BOOL debuggerEnabled;
extern MPSemaphoreID ss_sem;
extern BOOL ntscVideo;

/* from sms.c */
extern int sms_psg_enabled;
extern sn76489_t psg;
extern int sms_initialized;

int started = 0;

@implementation EmuController

#if MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3
- (void)application:(NSApplication *)app openFiles:(NSArray *)filenames
{
    NSString *fn;

    /* Eventually we should do something different if for some reason someone
       tries to send us multiple files, but for now, punt and open the first
       file in the list */
    fn = [filenames objectAtIndex:0];

    [memoryLock lock];

    if([self initMasterSystem] == NO)   {
        printf("could not init master system, bailing out!\n");
        exit(-1);
    }

    if([self loadMasterSystemRomFromFile:fn] == NO) {
        NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
            " file %@.", fn];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read from"
                        "this directory, and this file.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];

    [NSApp replyToOpenOrPrint:NSApplicationDelegateReplySuccess];
}

#else

- (BOOL)application:(NSApplication *)app openFile:(NSString *)filename
{
    [memoryLock lock];

    if([self initMasterSystem] == NO)   {
        printf("could not init master system, bailing out!\n");
        exit(-1);
    }

    if([self loadMasterSystemRomFromFile:filename] == NO) {
        NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read from"
                        "this directory, and this file.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];

    return YES;
}

#endif

-(NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app
{
    if(sms_initialized == 1)
        sms_write_cartram_to_file();

    return NSTerminateNow;
}

- (void)awakeFromNib
{
    prefs = [[NSUserDefaults standardUserDefaults] retain];

    openRom = nil;
    started = 0;

    if([prefs objectForKey: @"SMS Buttons"] != nil) {
        buttonmap = [NSString stringWithString: [prefs objectForKey: @"SMS Buttons"]];
        [buttonmap retain];
    }
    else    {
        unichar string[14];

        string[0] = 0x7E;
        string[1] = 0x7D;
        string[2] = 0x7B;
        string[3] = 0x7C;
        string[4] = 0x29;
        string[5] = 0x27;
        string[6] = 0x23;
        string[7] = 0x0D;
        string[8] = 0x01;
        string[9] = 0x00;
        string[10] = 0x02;
        string[11] = 0x09;
        string[12] = 0x0B;
        string[13] = 0x69;

        buttonmap = [NSString stringWithCharacters:string length:14];
        [buttonmap retain];
        
        [prefs setObject:buttonmap forKey:@"SMS Buttons"];
        [prefs synchronize];
    }
}

- (void)dealloc
{
    [openRom release];
    [prefs setObject:buttonmap forKey:@"SMS Buttons"];
    [prefs synchronize];
    [prefs release];
    [buttonmap release];

    [super dealloc];
}

- (IBAction)enableNoise:(id)sender
{
	if([noise state] == NSOnState)	{
		[noise setState: NSOffState];
		psg.enabled_channels &= ~ENABLE_NOISE;
	}
	else	{
		[noise setState: NSOnState];
		psg.enabled_channels |= ENABLE_NOISE;
	}
}

- (IBAction)enablePSG:(id)sender
{
	if([PSG state] == NSOnState)	{
		[PSG setState:NSOffState];
		sms_psg_enabled = 0;
	}
	else	{
		[PSG setState:NSOnState];
		sms_psg_enabled = 1;
	}
}

- (IBAction)enableTone0:(id)sender
{
	if([tone0 state] == NSOnState)	{
		[tone0 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE0;
	}
	else	{
		[tone0 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE0;
	}
}

- (IBAction)enableTone1:(id)sender
{
	if([tone1 state] == NSOnState)	{
		[tone1 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE1;
	}
	else	{
		[tone1 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE1;
	}
}

- (IBAction)enableTone2:(id)sender
{
	if([tone2 state] == NSOnState)	{
		[tone2 setState:NSOffState];
		psg.enabled_channels &= ~ENABLE_TONE2;
	}
	else	{
		[tone2 setState:NSOnState];
		psg.enabled_channels |= ENABLE_TONE2;
	}
}

- (IBAction)rendererNormal:(id)sender
{
    if([normalSize state] != NSOnState) {
        [normalSize setState:NSOnState];
        [doubleSize setState:NSOffState];
        [theview setDoubleSize:NO];

        gui_resized();
    }
}

- (IBAction)rendererDouble:(id)sender
{
    if([doubleSize state] != NSOnState) {
        [doubleSize setState:NSOnState];
        [normalSize setState:NSOffState];
        [theview setDoubleSize:YES];

        gui_resized();
    }
}

- (IBAction)loadMasterRom:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"sms", @"gg", nil];
	[openPanel setTitle:@"Choose a Master System/Game Gear Rom"];
	[openPanel setCanChooseDirectories:NO];
	
	if([openPanel runModalForTypes:fileTypes] == NSOKButton)	{
		NSString *fn = [openPanel filename];

        [memoryLock lock];

        if([self initMasterSystem] == NO)   {
            printf("could not init master system, bailing out!\n");
            exit(-1);
        }

        if([self loadMasterSystemRomFromFile:fn] == NO) {
            NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
                " file %@.", fn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                            "Verify that you have the permissions to read from"
                            "this directory, and this file.", 
                            @"OK", NULL, NULL);
        }

        [memoryLock unlock];
	}
}

- (IBAction)prefGGStart:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];
    
    buttons[6] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];
}

- (IBAction)prefSms1B1:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters:buttons];

    buttons[4] = [self runVirtKeyboard];

    [buttonmap release];

    buttonmap = [NSString stringWithCharacters:buttons length:14];
    [buttonmap retain];
}

- (IBAction)prefSms1B2:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[5] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms1Down:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[1] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms1Left:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[2] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms1Reset:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[0] = 0x7E;
    buttons[1] = 0x7D;
    buttons[2] = 0x7B;
    buttons[3] = 0x7C;
    buttons[4] = 0x29;
    buttons[5] = 0x27;
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}    

- (IBAction)prefSms1Right:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[3] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms1Up:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[0] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2B1:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[11] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2B2:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[12] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2Down:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[8] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2Left:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[9] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2Reset:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[7] = 0x0D;
    buttons[8] = 0x01;
    buttons[9] = 0x00;
    buttons[10] = 0x02;
    buttons[11] = 0x09;
    buttons[12] = 0x0B;
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2Right:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[10] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (IBAction)prefSms2Up:(id)sender
{
    unichar buttons[14];
    [buttonmap getCharacters: buttons];
    
    buttons[7] = [self runVirtKeyboard];
    
    [buttonmap release];
    
    buttonmap = [NSString stringWithCharacters: buttons length: 14];
    [buttonmap retain];
}

- (unichar)runVirtKeyboard
{
    unichar retval;

    [NSApp beginSheet: virtKeyboard modalForWindow: prefWindow modalDelegate: self didEndSelector: NULL contextInfo: NULL];

    retval = (unichar) [NSApp runModalForWindow: virtKeyboard];

    [NSApp endSheet: virtKeyboard];
    [virtKeyboard orderOut: nil];

    return retval;
}

- (void)endVirtKeyboard:(unichar)character
{
    [NSApp stopModalWithCode: (int)character];
}

- (void)masterEmuThread:(id)junk
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

	emu_loop();

	sms_shutdown();
	[pool release];
}

- (void)pauseEmulation:(id)sender
{
    if(started == 0)
        return;

    if([pause state] == NSOffState) {
        [pause setState: NSOnState];
        [memoryLock lock];
    }
    else    {
        [pause setState: NSOffState];
        [memoryLock unlock];
    }
}

- (IBAction)enableDebugger:(id)sender
{
    if(started == 0)
        return;

    if([debugger state] == NSOffState) {
        [debugger setState: NSOnState];
        [debuggerWindow makeKeyAndOrderFront: self];
        debuggerEnabled = YES;
    }
    else    {
        [debugger setState: NSOffState];
        [debuggerWindow orderOut: self];
        debuggerEnabled = NO;
        MPSignalSemaphore(ss_sem);
    }
}

- (void)updateButtonMapFromString: (NSString *)str
{
    [str retain];
    [buttonmap release];
    buttonmap = str;
}

- (NSString *)buttonMap
{
    return buttonmap;
}

- (IBAction)saveSRAM:(id)sender
{
    if(sms_initialized == 1)
        sms_write_cartram_to_file();
}

- (IBAction)saveState1:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".1.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState2:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".2.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState3:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".3.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState4:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".4.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState5:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".5.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)saveState6:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".6.cestate"];

    [memoryLock lock];

    if(sms_save_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to this "
                        "directory and that the target volume is not full.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState1:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".1.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState2:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".2.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState3:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".3.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState4:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".4.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState5:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".5.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)loadState6:(id)sender
{
    NSString *filename;

    if(sms_initialized == 0)
        return;

    filename = [openRom stringByAppendingString:@".6.cestate"];

    [memoryLock lock];

    if(sms_load_state([filename UTF8String]))   {
        NSString *title = [NSString stringWithFormat:@"Cannot open save state"
            " file %@.", filename];
        NSRunAlertPanel(title, @"CrabEmu cannot open this file for reading. "
                        "Verify that you have the permissions to read to this "
                        "directory and that the file exists.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)resetEmulation:(id)sender
{
    sms_soft_reset();
}

- (IBAction)modeNTSC:(id)sender
{
    if(sms_initialized != 1)    {
        if([ntsc state] == NSOffState) {
            [ntsc setState: NSOnState];
            [pal setState: NSOffState];
            ntscVideo = YES;
        }
    }
}

- (IBAction)modePAL:(id)sender
{
    if(sms_initialized != 1)    {
        if([pal state] == NSOffState) {
            [pal setState: NSOnState];
            [ntsc setState: NSOffState];
            ntscVideo = NO;
        }
    }
}

- (BOOL)initMasterSystem
{
    if(sms_initialized == 1)    {
        sound_reset_buffer();
        if(sms_reset())
            return NO;
    }
    else    {
        if(sms_init(ntscVideo == YES ? SMS_VIDEO_NTSC : SMS_VIDEO_PAL))
            return NO;
    }

    if([doubleSize state] == NSOnState) {
        gui_set_viewport(512, 384);
    }

    [theview showWindow];

    if(started == 0)    {
        [NSThread detachNewThreadSelector:@selector(masterEmuThread:)
                                 toTarget:self
                               withObject:nil];
        started = 1;
    }

    [ntsc setEnabled:NO];
    [pal setEnabled:NO];

    return YES;
}

- (BOOL)loadMasterSystemRomFromFile:(NSString *)filename
{
    if(openRom != nil)  {
        [openRom release];
    }

    openRom = [filename retain];

    if(sms_mem_load_rom([filename UTF8String])) {
        [openRom release];
        openRom = nil;

        return NO;
    }

    return YES;
}

@end
