/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <AudioUnit/AudioUnit.h>
#include <pthread.h>
#include "sound.h"

#define UNUSED __attribute__((unused))

static AudioUnit outputAU;
static unsigned char buffer[44100 * 20 * 2];
static UInt32 buffer_size = 16384;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static OSStatus sound_callback(void *inRefCon UNUSED, 
                               AudioUnitRenderActionFlags inActionFlags UNUSED,
                               const AudioTimeStamp *inTimeStamp UNUSED,
                               UInt32 inBusNumber UNUSED,
                               AudioBuffer *ioData) {
    UInt32 len = ioData->mDataByteSize;
    void *ptr = ioData->mData;

    pthread_mutex_lock(&mutex);

    if(buffer_size < len)   {
        pthread_mutex_unlock(&mutex);
        return noErr;
    }
    else    {
        memcpy(ptr, buffer, len);
        buffer_size -= len;
        
        if(buffer_size)
            memcpy(buffer, buffer + len, buffer_size);
    }

    pthread_mutex_unlock(&mutex);

    return noErr;
}

void sound_update_buffer(signed short *buf, int length) {
    pthread_mutex_lock(&mutex);

    memcpy(buffer + buffer_size, buf, length);
    buffer_size += length;

    pthread_mutex_unlock(&mutex);
}

int sound_init(void)    {
    OSStatus error = noErr;
    ComponentDescription desc;
    AudioStreamBasicDescription basic_desc;
    Component comp;
    struct AudioUnitInputCallback callback;
    UInt32 bufsz, size;
    AudioDeviceID dev;
    AudioTimeStamp timestamp;

    memset(buffer, 0, 44100 * 20 * 2);

    /* Set up the AudioStreamBasicDescription */
    basic_desc.mFormatID = kAudioFormatLinearPCM;
    basic_desc.mFormatFlags = kLinearPCMFormatFlagIsPacked |
                              kLinearPCMFormatFlagIsSignedInteger |
                              kAudioFormatFlagsNativeEndian;
    basic_desc.mChannelsPerFrame = 2;
    basic_desc.mSampleRate = 44100;
    basic_desc.mBitsPerChannel = 16;
    basic_desc.mFramesPerPacket = 1;
    basic_desc.mBytesPerFrame = 4;
    basic_desc.mBytesPerPacket = 4;

    /* And find the default audio output unit */
    desc.componentType = kAudioUnitComponentType;
    desc.componentSubType = kAudioUnitSubType_Output;
    desc.componentManufacturer = kAudioUnitID_DefaultOutput;
    desc.componentFlags = 0;
    desc.componentFlagsMask = 0;

    comp = FindNextComponent(NULL, &desc);

    if(comp == NULL)    {
        return -1;
    }

    error = OpenAComponent(comp, &outputAU);

    if(error != noErr)  {
        return -2;
    }

    error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_StreamFormat,
                                 kAudioUnitScope_Input, 0, &basic_desc,
                                 sizeof(basic_desc));

    if(error != noErr)  {
        return -3;
    }

    error = AudioUnitSetParameter(outputAU, kAudioUnitParameterUnit_LinearGain,
                                  kAudioUnitScope_Output, 0, 1.25f, 0);

    if(error != noErr)  {
        return -4;
    }

    callback.inputProc = sound_callback;
    callback.inputProcRefCon = NULL;

    error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_SetInputCallback,
                                 kAudioUnitScope_Input, 0, &callback,
                                 sizeof(callback));

    if(error != noErr)  {
        return -5;
    }

    size = sizeof(dev);

    error = AudioUnitGetProperty(outputAU,
                                 kAudioOutputUnitProperty_CurrentDevice,
                                 kAudioUnitScope_Global, 0, &dev, &size);

    if(error != noErr)  {
        return -6;
    }

    bufsz = 2048;
    timestamp.mFlags = 0;
    error = AudioDeviceSetProperty(dev, &timestamp, 0, false, 
                                   kAudioDevicePropertyBufferFrameSize,
                                   sizeof(UInt32), &bufsz);

    if(error != noErr)  {
        return -7;
    }

    error = AudioUnitSetProperty(outputAU,
                                 kAudioUnitProperty_MaximumFramesPerSlice,
                                 kAudioUnitScope_Global, 0, &bufsz,
                                 sizeof(UInt32));

    error = AudioUnitInitialize(outputAU);
    if(error != noErr)  {
        return -8;
    }

    error = AudioOutputUnitStart(outputAU);
    if(error != noErr)  {
        return -9;
    }

    return 0;
}

void sound_shutdown(void)   {
    OSStatus error;
    struct AudioUnitInputCallback callback;

    error = AudioOutputUnitStop(outputAU);
    if(error != noErr)
        return;

    callback.inputProc = NULL;
    callback.inputProcRefCon = NULL;

    error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_SetInputCallback,
                                 kAudioUnitScope_Input, 0, &callback,
                                 sizeof(callback));
    if(error != noErr)
        return;

    error = AudioUnitUninitialize(outputAU);
    if(error != noErr)
        return;

    error = CloseComponent(outputAU);
    if(error != noErr)
        return;
}

void sound_reset_buffer(void)   {
    memset(buffer, 0, 16384);
    buffer_size = 16384;
}
