/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.
 
    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "CrabZ80.h"
#include "CrabZ80_tables.h"
#include "CrabZ80_macros.h"
#include <stdio.h>
#include <assert.h>

static CrabZ80_t *cpu = NULL;

#ifndef CRABZ80_BIG_SWITCHES

typedef void (*opcodeptr)();

static const opcodeptr _op_tab[0x100];
static const opcodeptr _opCB_tab[0x100];
static const opcodeptr _opDFD_tab[0x100];
static const opcodeptr _opED_tab[0x100];
static const opcodeptr _opDFDCB_tab[0x100];

#define EXECUTE(prefix, op) {   \
    uint8 _tmp = (op); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    _op##prefix##_tab[_tmp](); \
}

#else

static inline void __CrabZ80_execute(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeCB(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeDFD(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeED(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeDFDCB(uint8 op, CrabZ80_t *cpu);

#define EXECUTE(prefix, op) {   \
    uint8 _tmp = (op); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    __CrabZ80_execute ##prefix (_tmp, cpu); \
}

#endif

#include "CrabZ80_opcodes.h"

static uint8 CrabZ80_dummy_read(uint16 addr __attribute__((unused)))    {
    return 0;
}

static void CrabZ80_dummy_write(uint16 addr __attribute__((unused)), 
                                uint8 data __attribute__((unused))) {
}

void CrabZ80_set_portread(CrabZ80_t *cpuz80, uint8 (*pread)(uint16 port))   {
    if(pread == NULL)
        cpuz80->pread = CrabZ80_dummy_read;
    else
        cpuz80->pread = pread;
}

void CrabZ80_set_memread(CrabZ80_t *cpuz80, uint8 (*mread)(uint16 addr))    {
    if(mread == NULL)
        cpuz80->mread = CrabZ80_dummy_read;
    else
        cpuz80->mread = mread;
}

void CrabZ80_set_portwrite(CrabZ80_t *cpuz80,
                           void (*pwrite)(uint16 port, uint8 data)) {
    if(pwrite == NULL)
        cpuz80->pwrite = CrabZ80_dummy_write;
    else
        cpuz80->pwrite = pwrite;
}

void CrabZ80_set_memwrite(CrabZ80_t *cpuz80,
                          void (*mwrite)(uint16 addr, uint8 data))  {
    if(mwrite == NULL)
        cpuz80->mwrite = CrabZ80_dummy_write;
    else
        cpuz80->mwrite = mwrite;
}

void CrabZ80_init(CrabZ80_t *cpuz80)    {
    cpuz80->pread = CrabZ80_dummy_read;
    cpuz80->mread = CrabZ80_dummy_read;
    cpuz80->pwrite = CrabZ80_dummy_write;
    cpuz80->mwrite = CrabZ80_dummy_write;
}

void CrabZ80_reset(CrabZ80_t *cpuz80)   {
    cpuz80->pc.w = 0x0000;
    cpuz80->iff1 = 0;
    cpuz80->iff2 = 0;
    cpuz80->im = 0;
    cpuz80->halt = 0;
    cpuz80->r_top = 0;
    cpuz80->af.w = 0x0040;
    cpuz80->bc.w = 0x0000;
    cpuz80->de.w = 0x0000;
    cpuz80->hl.w = 0x0000;
    cpuz80->ix.w = 0xFFFF;
    cpuz80->iy.w = 0xFFFF;
    cpuz80->sp.w = 0x0000;
    cpuz80->ir.w = 0x0000;
    cpuz80->afp.w = 0x0000;
    cpuz80->bcp.w = 0x0000;
    cpuz80->dep.w = 0x0000;
    cpuz80->hlp.w = 0x0000;
    cpuz80->cycles = 0;
    cpuz80->cycles_in = 0;
    cpuz80->ei = 0;
    cpuz80->irq_pending = 0;
}

void CrabZ80_pulse_nmi(CrabZ80_t *cpuz80)   {
    if(cpuz80->halt)    {
        cpuz80->pc.w++;
        cpuz80->halt = 0;
    }

    ++cpuz80->ir.b.l;

    cpuz80->iff1 = 0;
    cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.h);
    cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.l);
    cpuz80->pc.w = 0x0066;
    cpuz80->cycles += 11;

    cpuz80->ir.b.l = (cpuz80->ir.b.l & 0x80) | ((cpuz80->ir.b.l + 1) & 0x7F);
}

static void CrabZ80_take_irq(CrabZ80_t *cpuz80) {
    if(cpuz80->iff1 == 0) /* interrupts are disabled. */
        return;

    if(cpuz80->halt)    {
        cpuz80->pc.w++;
        cpuz80->halt = 0;
    }

    ++cpuz80->ir.b.l;

    cpuz80->iff1 = cpuz80->iff2 = 0;

    switch(cpuz80->im)  {
        /* This case 0 is a hack specific to the Sega Master System 2/Game Gear
           technically, we should have some sort of function to give us the
           opcode to use in mode 0, but its not implemented yet. If someone
           really wants to use CrabZ80 for something that needs mode 0 or
           mode 2, let me know, and I'll implement them properly. Until then,
           or when I decide to do it, it'll stay this way. */
        case 0:
        case 1:
            cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.h);
            cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.l);
            cpuz80->pc.w = 0x0038;
            cpuz80->cycles += 13;
            break;
        default:
            printf("Unimplemented Interrupt Mode: %d\n", cpuz80->im);
            assert(0);
    }
}

void CrabZ80_assert_irq(CrabZ80_t *cpuz80)  {
    cpuz80->irq_pending = 1;
}

void CrabZ80_clear_irq(CrabZ80_t *cpuz80)   {
    cpuz80->irq_pending = 0;
}

void CrabZ80_release_cycles()   {
    cpu->cycles_in = 0;
}

uint32 CrabZ80_execute(CrabZ80_t *cpuin, uint32 cycles) {
    uint8 inst;
    uint32 real_cycles;

    cpu = cpuin;

    cpu->cycles_in = cycles;

    while(cpu->cycles < cpu->cycles_in) {
        if(cpu->irq_pending && !cpu->ei)    {
            CrabZ80_take_irq(cpu);
        }

        cpu->ei = 0;
        inst = cpu->mread(cpu->pc.w++);
        EXECUTE(, inst);
        ++cpu->ir.b.l;
    }

    real_cycles = cpu->cycles;
    cpu->cycles = 0;

    cpu = NULL;

    return real_cycles;
}
