/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#import <Cocoa/Cocoa.h>
#include <CoreFoundation/CFDate.h>
#include "EmuOGLView.h"
#include "Z80DebugController.h"

#include <OpenGL/GL.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"

EmuOGLView *theview;
Z80DebugController *debugger;
NSLock *memoryLock;
BOOL debuggerEnabled = NO;
int debuggerMode = 0;
MPSemaphoreID ss_sem;
BOOL ntscVideo = YES;

static GLuint glTexture;
static BOOL glInitted = NO;
static BOOL resized = NO;
extern sms_vdp_t smsvdp;
extern int sms_initialized;
extern int sms_console;
static unsigned long frame_count = 0;

@interface SegaMasterSystem : NSObject
{
    NSTimer *frameTimer;
}

- (void)beginEmulation:(id)sender;
- (void)timerFired:(id)sender;

@end

@implementation SegaMasterSystem
- (void)beginEmulation:(id)sender
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    double denominator = ntscVideo == YES ? 60.0 : 50.0;

    frameTimer = [[NSTimer scheduledTimerWithTimeInterval:1.0 / denominator
                                                   target:self
                                                 selector:@selector(timerFired:)
                                                 userInfo:nil
                                                  repeats:YES] retain];

    [[NSRunLoop currentRunLoop] run];
    [pool release];
}

- (void)timerFired:(id)sender
{
    ++frame_count;    
}
@end

void render_scene(void) {
    if(!glInitted)  {
        if(!sms_initialized)
            return;

        glGenTextures(1, &glTexture);
        glBindTexture(GL_TEXTURE_2D, glTexture);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, smsvdp.framebuffer);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        glEnable(GL_TEXTURE_2D);

        glInitted = YES;
        resized = YES;
    }

    if(resized) {
        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(1.0f);

        if([theview doubleSize])    {
            glViewport(0.0f, 0.0f, [theview width] * 2, [theview height] * 2);
        }
        else    {
            glViewport(0.0f, 0.0f, [theview width], [theview height]);
        }

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        if(sms_console != CONSOLE_GG)
            glOrtho(0.0f, 256.0f, smsvdp.lines, 0.0f, 1.0f, 0.0f);
        else
            glOrtho(0.0f, 160.0f, 144.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        resized = NO;
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    glBindTexture(GL_TEXTURE_2D, glTexture);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 256, GL_RGBA, GL_UNSIGNED_BYTE, smsvdp.framebuffer);

    if(sms_console != CONSOLE_GG)   {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, smsvdp.lines / 256.0f);
        glVertex2f(0.0f, smsvdp.lines);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, smsvdp.lines / 256.0f);
        glVertex2f(256.0f, smsvdp.lines);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(256.0f, 0.0f);
        glEnd();
    }
    else    {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(48.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(0.0f, 144.0f);
        glTexCoord2f(48.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(208.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(160.0f, 144.0f);
        glTexCoord2f(208.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(160.0f, 0.0f);
        glEnd();
    }

    glFlush();
}

void emu_loop(void) {
#ifdef TIMING_TEST
    time_t thetime = time(NULL);
    unsigned long frames = 0;
#else
    CFAbsoluteTime timer = CFAbsoluteTimeGetCurrent();
    int hz = ntscVideo == YES ? 60 : 50;
#endif
    SegaMasterSystem *sms = [[SegaMasterSystem alloc] init];

    [NSThread detachNewThreadSelector:@selector(beginEmulation:)
                             toTarget:sms
                           withObject:nil];

    for(;;) {
#ifndef TIMING_TEST
        unsigned long old_frames = frame_count;
#endif

        if(debuggerEnabled == NO)   {
            [memoryLock lock];

            sms_frame();

            [theview display];

            [memoryLock unlock];

#ifndef TIMING_TEST
            while(frame_count == old_frames)
                usleep(1000);

            while(timer + 1.0 / hz > CFAbsoluteTimeGetCurrent())
                usleep(10);

            timer = CFAbsoluteTimeGetCurrent();
#else
            ++frames;

            if(thetime + 5 <= time(NULL))   {
                printf("%lu frames in %d seconds = %f FPS\n", frames, (int)(time(NULL) - thetime), (float) frames / (time(NULL) - thetime));
                thetime = time(NULL);
                frames = 0;
            }
#endif
        }
        else    {
            int line;
            int cycles;
            int lines = ntscVideo == YES ? 262 : 313;

            [memoryLock lock];

            for(line = 0; line < lines; ++line)   {
                for(cycles = 227; cycles > 0; cycles -= sms_z80_run(1)) {
                    if(debuggerEnabled == YES && debuggerMode == 0)  {
                        [debugger updateWithCycles:cycles scanline:line];
                        MPWaitOnSemaphore(ss_sem, kDurationForever);
                    }
                }

                sms_vdp_execute(line);

                if(debuggerEnabled == YES && debuggerMode == 1) {
                    [debugger updateWithCycles:227 scanline:line];
                    MPWaitOnSemaphore(ss_sem, kDurationForever);
                }
            }

            [theview display];
            [memoryLock unlock];

            if(debuggerEnabled == YES && debuggerMode == 2) {
                [debugger updateWithCycles:227 scanline:226];
                MPWaitOnSemaphore(ss_sem, kDurationForever);
            }
        }
    }
}

void gui_set_viewport(int w, int h) {
    [theview resizeWindowWithViewWidth: (float) w
                                height: (float) h];
        
    resized = YES;
}

void gui_resized()  {
    resized = YES;
}

int main(int argc, char *argv[])    {
    int retval;

    memoryLock = [[NSLock alloc] init];
    MPCreateSemaphore(1, 0, &ss_sem);

    retval = NSApplicationMain(argc,  (const char **) argv);

    MPDeleteSemaphore(ss_sem);
    [memoryLock release];
}
