/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "smsvcnt.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef _arch_dreamcast
#include <malloc.h>
#endif

/* The hardwired palette when used in a TMS9918 mode */
/* These constants are taken from Sean Young's TMS9918A Documentation */
#ifndef _arch_dreamcast
static const sms_vdp_color_t tms9918_pal[16] = {
    0xFF000000, /* Transparent */
    0xFF000000, /* Black */
    0xFF21C842, /* Medium Green */
    0xFF5EDC78, /* Light Green */
    0xFF5455ED, /* Dark Blue */
    0xFF7D76FC, /* Light Blue */
    0xFFD4524D, /* Dark Red */
    0xFF42EBF5, /* Cyan */
    0xFFFC5554, /* Medium Red */
    0xFFFF7978, /* Light Red */
    0xFFD4C154, /* Dark Yellow */
    0xFFE6CE80, /* Light Yellow */
    0xFF21B03B, /* Dark Green */
    0xFFC95BBA, /* Magenta */
    0xFFCCCCCC, /* Gray */
    0xFFFFFFFF, /* White */
};
#else
static const sms_vdp_color_t tms9918_pal[16] = {
    0x0000, /* Transparent */ 
    0x0000, /* Black */
    0x2648, /* Medium Green */
    0x5EEF, /* Light Green */
    0x52BD, /* Dark Blue */
    0x7BBF, /* Light Blue */
    0xD289, /* Dark Red */
    0x475E, /* Cyan */
    0xFAAA, /* Medium Red */
    0xFBCF, /* Light Red */
    0xD60A, /* Dark Yellow */
    0xE670, /* Light Yellow */
    0x2587, /* Dark Green */
    0xCAD7, /* Magenta */
    0xCE79, /* Gray */
    0xFFFF  /* White */
};
#endif

#define DRAW_COLOR(color, line, pixel) { \
    sms_vdp_color_t *_tmp = \
        (sms_vdp_color_t *)(smsvdp.framebuffer) + ((line) << 8) + (pixel); \
    *_tmp = color; \
}

extern int sms_console;
sms_vdp_t smsvdp;

#ifndef _arch_dreamcast
static void update_local_pal_sms(int num)   {
    /* Calculate the RGB888 color from the RGB222 color */
    smsvdp.pal[num] = ((smsvdp.cram[num] & 0x03) << 22) |
                      ((smsvdp.cram[num] & 0x0C) << 12) |
                      ((smsvdp.cram[num] & 0x30) << 2) |
                      (0xFF000000);
}

static void update_local_pal_gg(int num)    {
    /* Calculate the RGB888 color from the RGB444 color */
    smsvdp.pal[num >> 1] = ((smsvdp.cram[num] & 0x0F) << 20) |
                           ((smsvdp.cram[num] & 0xF0) << 8) |
                           ((smsvdp.cram[num + 1] & 0x0F) << 4) |
                           (0xFF000000);
}
#else
static void update_local_pal_sms(int num)   {
    /* Calculate the RGB565 color from the RGB222 color */
    smsvdp.pal[num] = ((smsvdp.cram[num] & 0x03) << 14) |
                      ((smsvdp.cram[num] & 0x0C) << 7) |
                      ((smsvdp.cram[num] & 0x30) >> 1);
}

static void update_local_pal_gg(int num)    {
    /* Calculate the RGB565 color from the RGB444 color */
    smsvdp.pal[num >> 1] = ((smsvdp.cram[num] & 0x0F) << 12) |
                           ((smsvdp.cram[num] & 0xF0) << 3) |
                           ((smsvdp.cram[num + 1] & 0x0F) << 1);
}
#endif

static inline void readjust_name_table()    {
    if(smsvdp.lines == 192) {
        smsvdp.name_table = smsvdp.vram + ((smsvdp.regs[2] & 0x0E) << 10);
    }
    else    {
        smsvdp.name_table = smsvdp.vram + ((smsvdp.regs[2] & 0x0C) << 10) +
            0x700;
    }
}

typedef void (*smsvdp_draw_func)(int line);

static void dummy_draw(int line __attribute__((unused)))    {
}

static smsvdp_draw_func bg_draw = &dummy_draw;
static smsvdp_draw_func spr_draw = &dummy_draw;

static const uint8 *vcnt_tab = vcnt_ntsc_192;

static void sms_vdp_update_cache(int pat)   {
    unsigned char *bitplane;
    int i, j;
    unsigned char pixel;
    unsigned char *tex[4];
    unsigned short tmp[4];
    unsigned long *ltmp[2];

    /* Determine where we should start converting */
    bitplane = smsvdp.vram + (pat << 5);

    tmp[0] = bitplane[0];
    tmp[1] = bitplane[1] << 1;
    tmp[2] = bitplane[2] << 2;
    tmp[3] = bitplane[3] << 3;

    ltmp[0] = (unsigned long *)tmp;
    ltmp[1] = ((unsigned long *)tmp) + 1;

    tex[0] = smsvdp.pattern[pat].texture[0] + 7;
    tex[1] = smsvdp.pattern[pat].texture[1];
    tex[2] = smsvdp.pattern[pat].texture[2] + 63;
    tex[3] = smsvdp.pattern[pat].texture[3] + 56;

    for(i = 0; i < 8; ++i)  {
        for(j = 0; j < 8; ++j)  {
            pixel = (tmp[0] & 0x01) | (tmp[1] & 0x02) | (tmp[2] & 0x04) |
                (tmp[3] & 0x08);

            *tex[0]-- = pixel;
            *tex[1]++ = pixel;
            *tex[2]-- = pixel;
            *tex[3]++ = pixel;

            *ltmp[0] >>= 1;
            *ltmp[1] >>= 1;
        }

        tex[0] += 16;
        tex[3] -= 16;

        bitplane += 4;

        tmp[0] = bitplane[0];
        tmp[1] = bitplane[1] << 1;
        tmp[2] = bitplane[2] << 2;
        tmp[3] = bitplane[3] << 3;
    }

    smsvdp.pattern[pat].dirty = 0;
}

#define DRAW_BACKGROUND(pixel) { \
    DRAW_COLOR(bg, line, (i << 3) + pixel); \
}

#define DRAW_FOREGROUND(pixel) { \
    DRAW_COLOR(fg, line, (i << 3) + pixel); \
}

static void sms_vdp_m0_draw_bg(int line)    {
    sms_vdp_color_t bg, fg;
    unsigned char *name_table;
    unsigned char *pattern_gen;
    unsigned char *color_table;
    unsigned char pixels, pattern, color;
    int i, row;

    name_table = &smsvdp.vram[(smsvdp.regs[2] & 0x0F) << 10];
    pattern_gen = &smsvdp.vram[(smsvdp.regs[4] & 0x07) << 11];
    color_table = &smsvdp.vram[smsvdp.regs[3] << 6];

    row = line >> 3;

    for(i = 0; i < 32; ++i) {
        pattern = name_table[(row << 5) + i];
        pixels = *(pattern_gen + (pattern << 3) + (line & 0x07));
        color = *(color_table + (pattern >> 3));

        if(color & 0x0F)
            bg = tms9918_pal[color & 0x0F];
        else
            bg = tms9918_pal[smsvdp.regs[7] & 0x0F];

        if(color >> 4)
            fg = tms9918_pal[color >> 4];
        else
            fg = tms9918_pal[smsvdp.regs[7] & 0x0F];;

        if(pixels & 0x80)
            DRAW_FOREGROUND(0)
        else
            DRAW_BACKGROUND(0)

        if(pixels & 0x40)
            DRAW_FOREGROUND(1)
        else
            DRAW_BACKGROUND(1)

        if(pixels & 0x20)
            DRAW_FOREGROUND(2)
        else
            DRAW_BACKGROUND(2)

        if(pixels & 0x10)
            DRAW_FOREGROUND(3)
        else
            DRAW_BACKGROUND(3)

        if(pixels & 0x08)
            DRAW_FOREGROUND(4)
        else
            DRAW_BACKGROUND(4)

        if(pixels & 0x04)
            DRAW_FOREGROUND(5)
        else
            DRAW_BACKGROUND(5)

        if(pixels & 0x02)
            DRAW_FOREGROUND(6)
        else
            DRAW_BACKGROUND(6)

        if(pixels & 0x01)
            DRAW_FOREGROUND(7)
        else
            DRAW_BACKGROUND(7)
    }
}

#undef DRAW_BACKGROUND
#undef DRAW_FOREGROUND

#define DRAW_BACKDROP(pixel) { \
    DRAW_COLOR(bd, line, (i * 6) + pixel + 8); \
}

#define DRAW_TEXT(pixel) { \
    DRAW_COLOR(tc, line, (i * 6) + pixel + 8); \
}

static void sms_vdp_m1_draw_bg(int line)    {
    sms_vdp_color_t tc, bd;
    unsigned char *name_table;
    unsigned char *pattern_gen;
    unsigned char pixels;
    int i, row;

    bd = tms9918_pal[smsvdp.regs[7] & 0x0F];
    tc = tms9918_pal[(smsvdp.regs[7] >> 4) & 0x0F];

    name_table = &smsvdp.vram[(smsvdp.regs[2] & 0x0F) << 10];
    pattern_gen = &smsvdp.vram[(smsvdp.regs[4] & 0x07) << 11];

    row = line >> 3;

    for(i = 0; i < 8; ++i)  {
        DRAW_COLOR(bd, line, i);
        DRAW_COLOR(bd, line, i + 248);
    }

    for(i = 0; i < 40; ++i) {
        pixels = *(pattern_gen + (name_table[row * 40 + i] << 3) + (line & 0x07));

        if(pixels & 0x80)
            DRAW_TEXT(0)
        else
            DRAW_BACKDROP(0)

        if(pixels & 0x40)
            DRAW_TEXT(1)
        else
            DRAW_BACKDROP(1)

        if(pixels & 0x20)
            DRAW_TEXT(2)
        else
            DRAW_BACKDROP(2)

        if(pixels & 0x10)
            DRAW_TEXT(3)
        else
            DRAW_BACKDROP(3)

        if(pixels & 0x08)
            DRAW_TEXT(4)
        else
            DRAW_BACKDROP(4)

        if(pixels & 0x04)
            DRAW_TEXT(5)
        else
            DRAW_BACKDROP(5)
    }
}

#undef DRAW_BACKDROP
#undef DRAW_TEXT

#define DRAW_BACKGROUND(pixel) { \
    DRAW_COLOR(bg, line, (i << 3) + pixel); \
}

#define DRAW_FOREGROUND(pixel) { \
    DRAW_COLOR(fg, line, (i << 3) + pixel); \
}

static void sms_vdp_m2_draw_bg(int line)    {
    sms_vdp_color_t bg, fg;
    unsigned char *name_table;
    unsigned char *pattern_gen;
    unsigned char *color_table;
    unsigned char pixels, color;
    int pattern;
    int i, row, mask, mask2;

    row = line >> 3;    
    mask = ((smsvdp.regs[4] & 0x03) << 8) | 0xFF;
    mask2 = ((smsvdp.regs[3] & 0x7F) << 3) | 0x07;

    name_table = &smsvdp.vram[(smsvdp.regs[2] & 0x0F) << 10];
    pattern_gen = &smsvdp.vram[(smsvdp.regs[4] & 0x04) << 11];
    color_table = &smsvdp.vram[(smsvdp.regs[3] & 0x80) << 6];

    for(i = 0; i < 32; ++i) {
        pattern = name_table[(row << 5) + i] + ((row & 0x18) << 5);
        pixels = *(pattern_gen + ((pattern & mask) << 3) + (line & 0x07));
        color = *(color_table + ((pattern & mask2) << 3) + (line & 0x07));

        if(color & 0x0F)
            bg = tms9918_pal[color & 0x0F];
        else
            bg = tms9918_pal[smsvdp.regs[7] & 0x0F];

        if(color >> 4)
            fg = tms9918_pal[color >> 4];
        else
            fg = tms9918_pal[smsvdp.regs[7] & 0x0F];

        if(pixels & 0x80)
            DRAW_FOREGROUND(0)
        else
            DRAW_BACKGROUND(0)

        if(pixels & 0x40)
            DRAW_FOREGROUND(1)
        else
            DRAW_BACKGROUND(1)

        if(pixels & 0x20)
            DRAW_FOREGROUND(2)
        else
            DRAW_BACKGROUND(2)

        if(pixels & 0x10)
            DRAW_FOREGROUND(3)
        else
            DRAW_BACKGROUND(3)

        if(pixels & 0x08)
            DRAW_FOREGROUND(4)
        else
            DRAW_BACKGROUND(4)

        if(pixels & 0x04)
            DRAW_FOREGROUND(5)
        else
            DRAW_BACKGROUND(5)

        if(pixels & 0x02)
            DRAW_FOREGROUND(6)
        else
            DRAW_BACKGROUND(6)

        if(pixels & 0x01)
            DRAW_FOREGROUND(7)
        else
            DRAW_BACKGROUND(7)
    }
}

#undef DRAW_BACKGROUND
#undef DRAW_FOREGROUND

#define DRAW_PIXEL(pixel) { \
    DRAW_COLOR(c, line, (i << 3) + pixel); \
}

static void sms_vdp_m3_draw_bg(int line)    {
    unsigned char *name_table;
    unsigned char *pattern_gen;
    int row, i;
    unsigned char pattern, color;
    sms_vdp_color_t c;

    name_table = &smsvdp.vram[(smsvdp.regs[2] & 0x0F) << 10];
    pattern_gen = &smsvdp.vram[(smsvdp.regs[4] & 0x07) << 11];

    row = line >> 3;

    for(i = 0; i < 32; ++i) {
        pattern = name_table[(row << 5) + i];

        if(!(line & 0x04))
            color = *(pattern_gen + (pattern << 3) + ((row & 0x03) << 1));
        else
            color = *(pattern_gen + (pattern << 3) + ((row & 0x03) << 1) + 1);

        if(color & 0x0F)
            c = tms9918_pal[color & 0x0F];
        else
            c = tms9918_pal[smsvdp.regs[7] & 0x0F];

        DRAW_PIXEL(0)
        DRAW_PIXEL(1)
        DRAW_PIXEL(2)
        DRAW_PIXEL(3)

        if(color >> 4)
            c = tms9918_pal[color & 0x0F];
        else
            c = tms9918_pal[smsvdp.regs[7] & 0x0F];

        DRAW_PIXEL(4)
        DRAW_PIXEL(5)
        DRAW_PIXEL(6)
        DRAW_PIXEL(7)
    }
}

#undef DRAW_PIXEL

#define CHECK_PIXEL(i)  { \
    if(col_tab[(i << size_shift)]) \
        smsvdp.status |= 0x20; \
\
    col_tab[(i << size_shift)] = 1; \
\
    if(size_shift)  { \
        if(col_tab[(i << 1) + 1]) \
            smsvdp.status |= 0x20; \
\
        col_tab[(i << 1) + 1] = 1; \
    } \
}

#define DRAW_PIXEL(i) { \
    DRAW_COLOR(c, line, (i << size_shift) + x); \
\
    if(size_shift) { \
        DRAW_COLOR(c, line, (i << 1) + x + 1); \
    } \
}

static void sms_vdp_m023_draw_spr(int line) {
    unsigned char *sat, *sprite_gen;
    int i, pattern_size, size_shift, tmp, pixels, pixels2;
    int count = 0, x, y, pattern, color;
    sms_vdp_color_t c;
    static unsigned char col_tab[0xFF];

    /* Grumble, grumble... shut up GCC */
    pixels2 = 0;

    /* First of all, clear out our colision table */
    memset(col_tab, 0, 0xFF);

    sat = &smsvdp.vram[(smsvdp.regs[5] & 0x7F) << 7];
    sprite_gen = &smsvdp.vram[(smsvdp.regs[6] & 0x07) << 11];

    if(smsvdp.regs[1] & 0x01)
        size_shift = 1;
    else
        size_shift = 0;

    if(smsvdp.regs[1] & 0x02)
        pattern_size = 16;
    else
        pattern_size = 8;

    for(i = 0; i < 32 && count < 5; ++i) {
        y = sat[i << 2] + 1;

        if(y == 0xD1)   {
            /* End of list marker */
            break;
        }
        else if(line < y)  {
            continue;
        }
        else if(y + (pattern_size << size_shift) - 1 < line)    {
            continue;
        }
        else if(++count == 5)   {
            break;
        }

        x = sat[(i << 2) + 1];
        pattern = sat[(i << 2) + 2];
        color = sat[(i << 2) + 3];

        if(color & 0x80)    {
            x -= 32;
        }

        tmp = (line - y) >> size_shift;

        if(pattern_size == 8)   {
            pixels = *(sprite_gen + (pattern << 3) + tmp);
        }
        else    {
            pixels = *(sprite_gen + ((pattern & 0xFC) << 3) + tmp);
            pixels2 = *(sprite_gen + ((pattern & 0xFC) << 3) + tmp + 16);
        }

        if(color & 0x0F)    {
            c = tms9918_pal[color & 0x0F];
        }
        else    {
            if(pattern_size == 8)   {
                if(pixels & 0x80 && x >= 0)
                    CHECK_PIXEL(0)

                if(pixels & 0x40 && x + 1 >= 0)
                    CHECK_PIXEL(1)

                if(pixels & 0x20 && x + 2 >= 0)
                    CHECK_PIXEL(2)

                if(pixels & 0x10 && x + 3 >= 0)
                    CHECK_PIXEL(3)

                if(pixels & 0x08 && x + 4 >= 0)
                    CHECK_PIXEL(4)

                if(pixels & 0x04 && x + 5 >= 0)
                    CHECK_PIXEL(5)

                if(pixels & 0x02 && x + 6 >= 0)
                    CHECK_PIXEL(6)

                if(pixels & 0x01 && x + 7 >= 0)
                    CHECK_PIXEL(7)
            }
            else    {
                if(pixels & 0x80 && x >= 0)
                    CHECK_PIXEL(0)

                if(pixels & 0x40 && x + 1 >= 0)
                    CHECK_PIXEL(1)

                if(pixels & 0x20 && x + 2 >= 0)
                    CHECK_PIXEL(2)

                if(pixels & 0x10 && x + 3 >= 0)
                    CHECK_PIXEL(3)

                if(pixels & 0x08 && x + 4 >= 0)
                    CHECK_PIXEL(4)

                if(pixels & 0x04 && x + 5 >= 0)
                    CHECK_PIXEL(5)

                if(pixels & 0x02 && x + 6 >= 0)
                    CHECK_PIXEL(6)

                if(pixels & 0x01 && x + 7 >= 0)
                    CHECK_PIXEL(7)

                if(pixels2 & 0x80 && x + 8 >= 0)
                    CHECK_PIXEL(8)

                if(pixels2 & 0x40 && x + 9 >= 0)
                    CHECK_PIXEL(9)

                if(pixels2 & 0x20 && x + 10 >= 0)
                    CHECK_PIXEL(10)

                if(pixels2 & 0x10 && x + 11 >= 0)
                    CHECK_PIXEL(11)

                if(pixels2 & 0x08 && x + 12 >= 0)
                    CHECK_PIXEL(12)

                if(pixels2 & 0x04 && x + 13 >= 0)
                    CHECK_PIXEL(13)

                if(pixels2 & 0x02 && x + 14 >= 0)
                    CHECK_PIXEL(14)

                if(pixels2 & 0x01 && x + 15 >= 0)
                    CHECK_PIXEL(15)
            }
            continue;
        }

        if(pattern_size == 8)   {
            if(pixels & 0x80 && x >= 0)
                DRAW_PIXEL(0)

            if(pixels & 0x40 && x + 1 >= 0)
                DRAW_PIXEL(1)

            if(pixels & 0x20 && x + 2 >= 0)
                DRAW_PIXEL(2)

            if(pixels & 0x10 && x + 3 >= 0)
                DRAW_PIXEL(3)

            if(pixels & 0x08 && x + 4 >= 0)
                DRAW_PIXEL(4)

            if(pixels & 0x04 && x + 5 >= 0)
                DRAW_PIXEL(5)

            if(pixels & 0x02 && x + 6 >= 0)
                DRAW_PIXEL(6)

            if(pixels & 0x01 && x + 7 >= 0)
                DRAW_PIXEL(7)
        }
        else    {
            if(pixels & 0x80 && x >= 0)
                DRAW_PIXEL(0)

            if(pixels & 0x40 && x + 1 >= 0)
                DRAW_PIXEL(1)

            if(pixels & 0x20 && x + 2 >= 0)
                DRAW_PIXEL(2)

            if(pixels & 0x10 && x + 3 >= 0)
                DRAW_PIXEL(3)

            if(pixels & 0x08 && x + 4 >= 0)
                DRAW_PIXEL(4)

            if(pixels & 0x04 && x + 5 >= 0)
                DRAW_PIXEL(5)

            if(pixels & 0x02 && x + 6 >= 0)
                DRAW_PIXEL(6)

            if(pixels & 0x01 && x + 7 >= 0)
                DRAW_PIXEL(7)

            if(pixels2 & 0x80 && x + 8 >= 0)
                DRAW_PIXEL(8)

            if(pixels2 & 0x40 && x + 9 >= 0)
                DRAW_PIXEL(9)

            if(pixels2 & 0x20 && x + 10 >= 0)
                DRAW_PIXEL(10)

            if(pixels2 & 0x10 && x + 11 >= 0)
                DRAW_PIXEL(11)

            if(pixels2 & 0x08 && x + 12 >= 0)
                DRAW_PIXEL(12)

            if(pixels2 & 0x04 && x + 13 >= 0)
                DRAW_PIXEL(13)

            if(pixels2 & 0x02 && x + 14 >= 0)
                DRAW_PIXEL(14)

            if(pixels2 & 0x01 && x + 15 >= 0)
                DRAW_PIXEL(15)
        }
    }

    if(!(smsvdp.status & 0x40))  {
        if(count == 5)  {
            /* Set the 5 sprites flag and the fifth sprite bits */
            smsvdp.status |= 0x40 | ((i - 1) & 0x1F);
        }
        else if(y == 0xD0)  {
            /* Set the fifth sprite bits to the last sprite displayed */
            smsvdp.status |= (i & 0x1F);
        }
        else    {
            /* Otherwise, set the fifth sprite bits to the last sprite */
            smsvdp.status |= 0x1F;
        }
    }
}

#undef DRAW_PIXEL
#undef CHECK_PIXEL

#define DRAW_PIXEL_HIGH()   { \
    entry = (*pixels++) + pal; \
    *current_line++ = smsvdp.pal[entry]; \
    *current_alpha++ = (entry & 0x0F); \
}

#define DRAW_PIXEL_LOW()    { \
    entry = (*pixels++) + pal; \
    *current_line++ = smsvdp.pal[entry]; \
    *current_alpha++ = 0; \
}

static void sms_vdp_m4_draw_bg(int line)    {
    int row, col, tile, tilenum, tmp;
    int rendercol, renderrow, xoff, yoff;
    unsigned char *pixels;
    int entry, pal, attrib;
    unsigned short *name_table_short;
    unsigned char *current_alpha;
    sms_vdp_color_t *current_line;

    row = line >> 3;

    /* Figure out what column/row to render first */
    if(row < 2 && smsvdp.regs[0] & 0x40)    {
        /* Do not scroll the top two rows, if requested. */
        rendercol = 0;
        xoff = 0;
    }
    else    {
        rendercol = smsvdp.xscroll_coarse;
        xoff = smsvdp.xscroll_fine;
    }

    if(smsvdp.lines == 192) {
        tmp = line + smsvdp.regs[9];

        if(tmp > 223)
            tmp -= 224;
    }
    else    {
        tmp = (line + smsvdp.regs[9]) & 0xFF;
    }

    yoff = smsvdp.yscroll_fine;

    renderrow = (tmp >> 3);

    name_table_short = &((unsigned short *)smsvdp.name_table)[renderrow << 5];

    current_line = (sms_vdp_color_t *)(smsvdp.framebuffer) + (line << 8) + xoff;
    current_alpha = smsvdp.alpha + xoff;

    /* Render each pixel of background data for this line */
    for(col = 0; col < 32; ++col)   {
        if(col == 24 && smsvdp.regs[0] & 0x80)  {
            /* Disable vertical scrolling for the last few columns */
            renderrow = line >> 3;
            yoff = 0;
            name_table_short = &((unsigned short *)smsvdp.name_table)[renderrow << 5];
        }

        /* Get the tile that we're rendering */
        tile = name_table_short[rendercol];
#ifdef __BIG_ENDIAN__
        tile = ((tile & 0xFF) << 8) | ((tile >> 8) & 0xFF);
#endif
        tilenum = tile & 0x1FF;

        /* Check if the pattern is dirtied, if so, clean it up */
        if(smsvdp.pattern[tilenum].dirty)
            sms_vdp_update_cache(tilenum);

        attrib = (tile >> 9) & 0x03;

        /* Find the texture */
        pixels = &smsvdp.pattern[tilenum].texture[attrib]
            [((line + yoff) & 0x07) << 3];

        pal = (tile & 0x800) >> 7;

        /* Draw the pixels */
        if(tile & 0x1000)   {
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
            DRAW_PIXEL_HIGH();
        }
        else    {
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
            DRAW_PIXEL_LOW();
        }

        rendercol = (rendercol + 1) & 0x1F;
    }
}

#undef DRAW_PIXEL_HIGH
#undef DRAW_PIXEL_LOW

static void sms_vdp_m4_draw_spr(int line)   {
    unsigned char *sat = smsvdp.sat;
    unsigned char *pix;
    static unsigned char col_tab[256];
    int height, i, n = 0, j, x, y, num, half, ds = 0, tiled = 0, ls = 0;
    sms_vdp_color_t *pal = (smsvdp.pal + 0x10);
    sms_vdp_color_t *current_pixel;

    /* First of all, clear out our colision table */
    memset(col_tab, 0, 0xFF);

    /* Normal sprites are 8 pixels high, so start with that */
    height = 8;

    /* If bit 1 of register 1 is set, then the height is 16 */
    if(smsvdp.regs[1] & 0x02)   {
        tiled = 1;
        height = 16;
    }

    /* If bit 0 of register 1 is set, sprites have their pixels doubled */
    if(smsvdp.regs[1] & 0x01)   {
        ds = 1;
        height <<= 1;
    }

    /* If bit 3 of register 0 is set, shift sprites left 8 pixels */
    ls = smsvdp.regs[0] & 0x08;

    /* Which half of the pattern table are we getting the patterns from? */
    half = (smsvdp.regs[6] & 0x04) << 6;

    /* Parse the table */
    for(i = 0; i < 64; ++i) {
        /* Check for the end of list flag */
        if(sat[i] == 0xD0)  {
            break;
        }

        y = sat[i] + 1;

        /* Check the position of this sprite */
        if(line >= y && line <= y + height - 1) {
            /* If its on this line, make sure that we haven't already rendered 8 sprites */
            if(n < 8)   {
                /* Determine the X position, which is stored in sat[i * 2 + 0x80], and can
                   be changed by bit 3 of register 0, also get the pattern number, which
                   is stored in sat[i * 2 + 0x81] */
                x = sat[(i << 1) | 0x80] - ls;
                num = sat[(i << 1) | 0x81] + half;

                if(tiled)
                    num &= 0x1FE;

                /* Set our pixel pointer up */
                if(((line - y) >> ds) > 7)  {
                    if(smsvdp.pattern[num + 1].dirty)
                        sms_vdp_update_cache(num + 1);
                    pix = &smsvdp.pattern[num + 1].texture[0][(((line - y) >>
                                                                ds) - 8) << 3];
                }
                else    {
                    if(smsvdp.pattern[num].dirty)
                        sms_vdp_update_cache(num);
                    pix = &smsvdp.pattern[num].texture[0][((line - y) >> ds) <<
                                                          3];
                }

                /* Don't draw to a negative coordinate */
                j = 0;
                if(x < 0)   {
                    j = 0 - x;
                    x = 0;
                }

                current_pixel = (sms_vdp_color_t *)(smsvdp.framebuffer) +
                    (line << 8) + x;

                for(; j < 8 && x < 256; ++j, ++x, ++current_pixel)  {
                    /* Check for collision, set the flag if there is one */
                    if(col_tab[x] == 1) {
                        smsvdp.status |= 0x20;
                        goto doublesize;
                    }

                    /* Make sure not to tromp on a high-priority background
                       tile */
                    if(smsvdp.alpha[x])
                        goto doublesize;

                    /* Actually draw the pixel */
                    if(pix[j])  {
                        *current_pixel = pal[pix[j]];
                        col_tab[x] = 1;
                    }

doublesize:
                    if(ds)  {
                        ++current_pixel;

                        /* Check for collision, set the flag if there is one */
                        if(col_tab[++x] == 1) {
                            smsvdp.status |= 0x20;
                            continue;
                        }
                        
                        /* Make sure not to tromp on a high-priority background
                           tile */
                        if(smsvdp.alpha[x])
                            continue;
                        
                        /* Actually draw the pixel */
                        if(pix[j])  {
                            *current_pixel = pal[pix[j]];
                            col_tab[x] = 1;
                        }
                    }
                }
                ++n;
            }
            else    {
                /* Too many sprites on this line, set the flag, and stop
                   rendering */
                smsvdp.status |= 0x40;
                break;
            }
        }
    }
}

void sms_vdp_execute(int line)  {
    int i;

    /* Draw only if the display is enabled */
    if(smsvdp.regs[0x1] & 0x40 && line < smsvdp.lines)  {
        bg_draw(line);
        spr_draw(line);

        /* Determine if we should mask column 0, and if so, do it */
        if(smsvdp.regs[0] & 0x20)   {
            int tmp = (smsvdp.regs[7] & 0x0F) | 0x10;
            sms_vdp_color_t *current_line =
                (sms_vdp_color_t *)(smsvdp.framebuffer) + (line << 8);

            for(i = 0; i < 8; ++i)  {
                *current_line++ = smsvdp.pal[tmp];
            }
        }
    }
    else if(line < smsvdp.lines)    {
        int tmp = (smsvdp.regs[7] & 0x0F) | 0x10;
        sms_vdp_color_t *current_line =
            (sms_vdp_color_t *)(smsvdp.framebuffer) + (line << 8);

        for(i = 0; i < 256; ++i)    {
            *current_line++ = smsvdp.pal[tmp];
        }
    }

    if(line < smsvdp.lines + 1) {
        if(smsvdp.linecnt == 0) {
            /* The Register 10 line count will underflow, interrupt if
               enabled */
            smsvdp.linecnt = smsvdp.regs[0xA];
            if(smsvdp.regs[0x0] & 0x10) {
                smsvdp.flags |= 0x04;
                sms_z80_run(1);
                sms_z80_assert_irq();
            }
        }
        else    {
            /* Decrement the line counter */
            --smsvdp.linecnt;
        }
    }
    else    {
        /* Reload the line counter */
        smsvdp.linecnt = smsvdp.regs[0xA];
    }

    if(line == smsvdp.lines + 1)    {
        /* Update the vertical scroll value, if needed */
        if(smsvdp.flags & 0x02) {
            smsvdp.regs[0x09] = smsvdp.v_scroll;

            smsvdp.yscroll_fine = smsvdp.v_scroll & 0x07;
            smsvdp.flags &= (~0x02);
        }

        /* Set the vblank flag, and signal an IRQ, if desired */
        smsvdp.status |= 0x80;

        if(smsvdp.regs[0x1] & 0x20)	{
            /* This should make Chicago Syndicate work. */
            sms_z80_run(1);
            sms_z80_assert_irq();
        }
    }

    smsvdp.line = line;
}

void sms_vdp_data_write(unsigned char data) {
    /* Clear the bytes written flag */
    smsvdp.flags &= (~0x01);

    /* First, update the read buffer */
    smsvdp.read_buf = data;

    /* Check the destination of this byte */
    switch(smsvdp.code) {
        case 0x00:
        case 0x01:
        case 0x02:
            smsvdp.vram[smsvdp.addr] = data;
            smsvdp.pattern[smsvdp.addr >> 5].dirty = 1;
            break;
        case 0x03:
            if(sms_console != CONSOLE_GG)   {
                smsvdp.cram[smsvdp.addr & 0x1F] = data;
                update_local_pal_sms(smsvdp.addr & 0x1F);
            }
            else    {
                if(!(smsvdp.addr & 0x01))   {
                    smsvdp.pal_latch = data;
                }
                else    {
                    smsvdp.cram[smsvdp.addr & 0x3E] = smsvdp.pal_latch;
                    smsvdp.cram[smsvdp.addr & 0x3F] = data;
                    update_local_pal_gg(smsvdp.addr & 0x3E);
                }
            }
            break;
    }

    /* Update the address register, and wrap, if needed */
    smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;
}

void sms_vdp_ctl_write(unsigned char data)  {
    /* First, update the code/address registers */
    if(smsvdp.flags & 0x01) {
        /* One byte written, write the last one */
        smsvdp.addr = (data & 0x3F) << 8 | (smsvdp.addr_latch & 0xFF);
        smsvdp.code = (data & 0xC0) >> 6;
        smsvdp.flags &= (~0x01);
    }
    else    {
        /* A multiple of 2 bytes written, start over */
        smsvdp.addr = (smsvdp.addr & 0x3F00) | (data);
        smsvdp.addr_latch = data;
        smsvdp.flags |= 0x01;
    }

    /* Now, handle register writes, if needed */
    if(!(smsvdp.flags & 0x01) && smsvdp.code == 0x02)   {
        int reg = (smsvdp.addr >> 8) & 0x0F;

        if(reg == 0x09 && smsvdp.line < smsvdp.lines)   {
            smsvdp.v_scroll = (smsvdp.addr & 0xFF);
            smsvdp.flags |= 0x02;
        }
        else    {
            smsvdp.regs[reg] = (smsvdp.addr & 0xFF);

            if(reg == 0)    {
                sms_vdp_set_vidmode(smsvdp.vidmode, smsvdp.machine);
            }
            else if(reg == 1)   {
                sms_vdp_set_vidmode(smsvdp.vidmode, smsvdp.machine);

                if((smsvdp.status & 0x80) && (smsvdp.regs[1] & 0x20))   {
                    sms_z80_run(1);
                    sms_z80_assert_irq();
                }
                else if((smsvdp.status & 0x80) && !(smsvdp.flags & 0x04))   {
                    /* Don't clear the IRQ line if there is a line interrupt
                       still going */
                    sms_z80_clear_irq();
                }
            }
            else if(reg == 2)   {
                readjust_name_table();
            }
            else if(reg == 5)   {
                smsvdp.sat = smsvdp.vram + ((smsvdp.regs[5] & 0x7E) << 7);
            }
            else if(reg == 8)   {
                int tmp = (smsvdp.regs[8] & 0xF8) >> 3;
                smsvdp.xscroll_coarse = (32 - tmp) & 0x1F;
                smsvdp.xscroll_fine = smsvdp.regs[8] & 0x07;
            }
            else if(reg == 9)   {
                smsvdp.yscroll_fine = smsvdp.regs[9] & 0x07;
            }
        }
    }

    /* Code 0 - Read a byte of vram, put it in the buffer, and increment
       the address (that last part is important!) */
    if(!(smsvdp.flags & 0x01) && smsvdp.code == 0x00)   {
        smsvdp.read_buf = smsvdp.vram[smsvdp.addr];
        smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;
    }
}

unsigned char sms_vdp_hcnt_read(void)   {
    /* This will always return 0, for the time being */
    /* return (unsigned char)(smsvdp.hcnt >> 1);*/
    return (unsigned char) 0;
}

unsigned char sms_vdp_vcnt_read(void)   {
    return vcnt_tab[smsvdp.line];
}

unsigned char sms_vdp_data_read(void)   {
    unsigned char tmp;

    /* Clear the bytes written flag */
    smsvdp.flags &= (~0x01);

    /* Fetch what is already in the buffer */
    tmp = smsvdp.read_buf;

    /* Fetch a new byte for the buffer */
    smsvdp.read_buf = smsvdp.vram[smsvdp.addr];

    /* And finally, increment the address */
    smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;

    return tmp;
}

unsigned char sms_vdp_status_read(void) {
    unsigned char tmp;

    /* Check the interrupt flags */
    if((smsvdp.status & 0x80) || (smsvdp.flags & 0x04)) {
        sms_z80_clear_irq();
    }

    /* Clear the bytes written flag and line interrupt pending flag */
    smsvdp.flags &= (~0x05);

    /* Fetch our flags, so that we can clear stale ones */
    tmp = smsvdp.status;
    smsvdp.status &= (~0xE0);

    return tmp;
}

int sms_vdp_init(int mode)  {
    int i, tmp;

    /* Initialize the VDP emulation to a sane state */
    smsvdp.code = 0;
    smsvdp.addr = 0;
    smsvdp.read_buf = 0;
    smsvdp.status = 0;
    smsvdp.hcnt = 0;
    smsvdp.vcnt = 0;
    smsvdp.flags = 0;
    smsvdp.linecnt = 0xFF;
    smsvdp.pal_latch = 0;

    /* Set some sane register values */
    smsvdp.regs[0x0] = 0x04;
    smsvdp.regs[0x1] = 0x00;
    smsvdp.regs[0x2] = 0x0E;
    smsvdp.regs[0x3] = 0x00;
    smsvdp.regs[0x4] = 0x00;
    smsvdp.regs[0x5] = 0x7F;
    smsvdp.regs[0x6] = 0x00;
    smsvdp.regs[0x7] = 0x00;
    smsvdp.regs[0x8] = 0x00;
    smsvdp.regs[0x9] = 0x00;
    smsvdp.regs[0xA] = 0x00;

    readjust_name_table();
    smsvdp.sat = smsvdp.vram + ((smsvdp.regs[5] & 0x7E) << 7);

    tmp = (smsvdp.regs[8] & 0xF8) >> 3;
    smsvdp.xscroll_coarse = (32 - tmp) & 0x1F;
    smsvdp.xscroll_fine = smsvdp.regs[8] & 0x07;
    smsvdp.yscroll_fine = smsvdp.regs[9] & 0x07;

    /* Mark all patterns as dirty */
    for(i = 0; i < 512; ++i)    {
        smsvdp.pattern[i].dirty = 1;
        /* There's no use in clearing the textures, since the will be
           overwritten before they're used, anyway */
    }

    /* Allocate memory */
    smsvdp.cram = (unsigned char *)malloc(64);

    if(smsvdp.cram == NULL) {
        fprintf(stderr, "sms_vdp_init: Could not allocate CRAM!\n");
        return -1;
    }

    smsvdp.pal = (sms_vdp_color_t *)malloc(sizeof(sms_vdp_color_t) * 32);

    if(smsvdp.pal == NULL)  {
        fprintf(stderr, "sms_vdp_init: Could not allocate local Palette!\n");
        free(smsvdp.cram);
        return -1;
    }

    smsvdp.vram = (unsigned char *)malloc(0x4000);

    if(smsvdp.vram == NULL) {
        fprintf(stderr, "sms_vdp_init: Could not allocate VRAM!\n");
        free(smsvdp.cram);
        free(smsvdp.pal);
        return -1;
    }

#ifndef _arch_dreamcast
    smsvdp.framebuffer = (unsigned char *)malloc(256 * 256 *
                                                 sizeof(sms_vdp_color_t));
#else
    smsvdp.framebuffer = (unsigned char *)memalign(32, 256 * 256 * 2);
#endif

    if(smsvdp.framebuffer == NULL)  {
        fprintf(stderr, "sms_vdp_init: Could not allocate framebuffer!\n");
        free(smsvdp.cram);
        free(smsvdp.pal);
        free(smsvdp.vram);
        return -1;
    }

#ifdef _arch_dreamcast
    smsvdp.framebuffer =
        (unsigned char *)((uint32)smsvdp.framebuffer | 0xA0000000);
#endif

    sms_vdp_set_vidmode(mode, SMS_VDP_MACHINE_SMS2);

    return 0;
}

int sms_vdp_reset(void) {
    int i;

    /* Initialize the VDP emulation to a sane state */
    smsvdp.code = 0;
    smsvdp.addr = 0;
    smsvdp.read_buf = 0;
    smsvdp.status = 0;
    smsvdp.hcnt = 0;
    smsvdp.vcnt = 0;
    smsvdp.flags = 0;
    smsvdp.linecnt = 0xFF;
    smsvdp.pal_latch = 0;

    /* Set some sane register values */
    smsvdp.regs[0x0] = 0x04;
    smsvdp.regs[0x1] = 0x00;
    smsvdp.regs[0x2] = 0x0E;
    smsvdp.regs[0x3] = 0x00;
    smsvdp.regs[0x4] = 0x00;
    smsvdp.regs[0x5] = 0x7F;
    smsvdp.regs[0x6] = 0x00;
    smsvdp.regs[0x7] = 0x00;
    smsvdp.regs[0x8] = 0x00;
    smsvdp.regs[0x9] = 0x00;
    smsvdp.regs[0xA] = 0x00;

    /* Mark all patterns as dirty */
    for(i = 0; i < 512; ++i)    {
        smsvdp.pattern[i].dirty = 1;
        /* There's no use in clearing the textures, since the will be
        overwritten before they're used, anyway */
    }

    memset(smsvdp.vram, 0, 0x4000);
    memset(smsvdp.cram, 0, 64);
    memset(smsvdp.pal, 0, sizeof(sms_vdp_color_t) * 32);

    sms_vdp_set_vidmode(smsvdp.vidmode, smsvdp.machine);

    return 0;
}

int sms_vdp_shutdown(void)  {
#ifdef _arch_dreamcast
    smsvdp.framebuffer =
        (unsigned char *)((uint32)smsvdp.framebuffer & 0xDFFFFFFF);
#endif

    /* Free all RAM used */
    free(smsvdp.cram);
    free(smsvdp.pal);
    free(smsvdp.vram);
    free(smsvdp.framebuffer);

    return 0;
}

void sms_vdp_set_vidmode(int mode, int machine) {
    int oldlines = smsvdp.lines;
    int vdp_mode;

    vdp_mode = ((smsvdp.regs[0x0] & 0x04) << 1) |
        ((smsvdp.regs[0x1] & 0x08) >> 1) |
        ((smsvdp.regs[0x0] & 0x02)) |
        ((smsvdp.regs[0x1] & 0x10) >> 4);

    switch(vdp_mode)    {
        case 0x00: /* Mode 0: TMS9918 Graphic 1 Mode */
            bg_draw = &sms_vdp_m0_draw_bg;
            spr_draw = &sms_vdp_m023_draw_spr;
            break;

        case 0x01: /* Mode 1: TMS9918 Text mode */
            bg_draw = &sms_vdp_m1_draw_bg;
            /* The sprite system is inactive */
            spr_draw = &dummy_draw;
            break;

        case 0x02: /* Mode 2: TMS9918 Graphic 2 Mode */
            bg_draw = &sms_vdp_m2_draw_bg;
            spr_draw = &sms_vdp_m023_draw_spr;
            break;

        case 0x04: /* Mode 3: TMS9918 Multicolor mode */
            bg_draw = &sms_vdp_m3_draw_bg;
            spr_draw = &sms_vdp_m023_draw_spr;
            break;

        case 0x08: /* Mode 4 */
        case 0x0A:
        case 0x0B:
        case 0x0C:
        case 0x0E:
        case 0x0F:
            bg_draw = &sms_vdp_m4_draw_bg;
            spr_draw = &sms_vdp_m4_draw_spr;
            break;

        default:
            fprintf(stderr, "sms_vdp_set_vidmode: Unsupported mode 0x%02x.\n", vdp_mode);
    }

    if(mode == SMS_VIDEO_NTSC)  {
        smsvdp.vidmode = SMS_VIDEO_NTSC;

        if(machine == SMS_VDP_MACHINE_SMS2) {
            smsvdp.machine = SMS_VDP_MACHINE_SMS2;

            /* Check the M1, M2, M3, and M4 bits for the video mode */
            if((smsvdp.regs[0] & 0x06) == 0x06)   {

                /* M4 and M2 are set: M1/M3 determine resolution */
                if(!(smsvdp.regs[1] & 0x18) ||
                   (smsvdp.regs[1] & 0x18) == 0x18)    {
                    /* Either both of M1/M3 are set, or neither is set:
                       192 line mode */
                    vcnt_tab = vcnt_ntsc_192;
                    smsvdp.lines = 192;
                }
                else if(smsvdp.regs[1] & 0x10)  {
                    /* M1 is set: 224 line mode */
                    vcnt_tab = vcnt_ntsc_224;
                    smsvdp.lines = 224;
                }
                else if(smsvdp.regs[1] & 0x08)  {
                    /* M3 is set: 240 line mode */
                    vcnt_tab = vcnt_ntsc_240;
                    smsvdp.lines = 240;
                }
                else    {
                    /* Invalid text mode.... */
                }
            }
            else if(smsvdp.regs[0] & 0x04)  {
                /* M4 is set */
                if(smsvdp.regs[1] & 0x10)   {
                    /* M1 is set: Invalid text mode */
                }
                else    {
                    /* M1 is not set: 192 line mode */
                    vcnt_tab = vcnt_ntsc_192;
                    smsvdp.lines = 192;
                }
            }
            else    {
                /* M4 is not set, use TMS9918 modes */
                vcnt_tab = vcnt_ntsc_192;
                smsvdp.lines = 192;
            }

            if(sms_console != CONSOLE_GG)
                gui_set_viewport(256, smsvdp.lines);
            else
                gui_set_viewport(160, 144);
        }
        else if(machine == SMS_VDP_MACHINE_SMS1)    {
        }
        else    {
            fprintf(stderr, "sms_vdp_set_vidmode: Invalid machine: %d\n",
                    machine);
        }
    }
    else if(mode == SMS_VIDEO_PAL)  {
        smsvdp.vidmode = SMS_VIDEO_PAL;

        if(machine == SMS_VDP_MACHINE_SMS2) {
            smsvdp.machine = SMS_VDP_MACHINE_SMS2;

            /* Check the M1, M2, M3, and M4 bits for the video mode */
            if((smsvdp.regs[0] & 0x06) == 0x06)   {
                
                /* M4 and M2 are set: M1/M3 determine resolution */
                if(!(smsvdp.regs[1] & 0x18) ||
                   (smsvdp.regs[1] & 0x18) == 0x18)    {
                    /* Either both of M1/M3 are set, or neither is set:
                       192 line mode */
                    vcnt_tab = vcnt_pal_192;
                    smsvdp.lines = 192;
                }
                else if(smsvdp.regs[1] & 0x10)  {
                    /* M1 is set: 224 line mode */
                    vcnt_tab = vcnt_pal_224;
                    smsvdp.lines = 224;
                }
                else if(smsvdp.regs[1] & 0x08)  {
                    /* M3 is set: 240 line mode */
                    vcnt_tab = vcnt_pal_240;
                    smsvdp.lines = 240;
                }
                else    {
                    /* Invalid text mode.... */
                }
            }
            else if(smsvdp.regs[0] & 0x04)  {
                /* M4 is set */
                if(smsvdp.regs[1] & 0x10)   {
                    /* M1 is set: Invalid text mode */
                }
                else    {
                    /* M1 is not set: 192 line mode */
                    vcnt_tab = vcnt_pal_192;
                    smsvdp.lines = 192;
                }
            }
            else    {
                /* M4 is not set, use TMS9918 modes */
                vcnt_tab = vcnt_pal_192;
                smsvdp.lines = 192;
            }

            if(sms_console != CONSOLE_GG)
                gui_set_viewport(256, smsvdp.lines);
            else
                gui_set_viewport(160, 144);
        }
        else if(machine == SMS_VDP_MACHINE_SMS1)    {
        }
        else    {
            fprintf(stderr, "sms_vdp_set_vidmode: Invalid machine: %d\n",
                    machine);
        }
    }
    else    {
        fprintf(stderr, "sms_vdp_set_vidmode: Invalid mode %d\n", mode);
    }

    if(oldlines != smsvdp.lines)    {
        readjust_name_table();
    }
}

void sms_vdp_write_context(FILE *fp)    {
    unsigned char byte;

    fwrite(&smsvdp.code, 1, 1, fp);

    byte = smsvdp.addr & 0xFF;
    fwrite(&byte, 1, 1, fp);

    byte = (smsvdp.addr >> 8) & 0xFF;
    fwrite(&byte, 1, 1, fp);

    fwrite(&smsvdp.addr_latch, 1, 1, fp);
    fwrite(&smsvdp.read_buf, 1, 1, fp);
    fwrite(smsvdp.cram, 64, 1, fp);
    fwrite(&smsvdp.status, 1, 1, fp);
    fwrite(smsvdp.regs, 16, 1, fp);
    fwrite(smsvdp.vram, 0x4000, 1, fp);
    fwrite(&smsvdp.vcnt, 1, 1, fp);
    fwrite(&smsvdp.pal_latch, 1, 1, fp);
    fwrite(&smsvdp.linecnt, 1, 1, fp);

    byte = smsvdp.flags & 0xFF;
    fwrite(&byte, 1, 1, fp);

    byte = (smsvdp.flags >> 8) & 0xFF;
    fwrite(&byte, 1, 1, fp);

    byte = (smsvdp.flags >> 16) & 0xFF;
    fwrite(&byte, 1, 1, fp);

    byte = (smsvdp.flags >> 24) & 0xFF;
    fwrite(&byte, 1, 1, fp);
}

void sms_vdp_read_context(FILE *fp)    {
    unsigned char byte[4];
    int i;

    fread(&smsvdp.code, 1, 1, fp);

    fread(&byte, 2, 1, fp);
    smsvdp.addr = byte[0] | (byte[1] << 8);

    fread(&smsvdp.addr_latch, 1, 1, fp);
    fread(&smsvdp.read_buf, 1, 1, fp);
    fread(smsvdp.cram, 64, 1, fp);
    fread(&smsvdp.status, 1, 1, fp);
    fread(smsvdp.regs, 16, 1, fp);
    fread(smsvdp.vram, 0x4000, 1, fp);
    fread(&smsvdp.vcnt, 1, 1, fp);
    fread(&smsvdp.pal_latch, 1, 1, fp);
    fread(&smsvdp.linecnt, 1, 1, fp);

    fread(&byte, 4, 1, fp);
    smsvdp.flags = byte[0] | (byte[1] << 8) | (byte[2] << 16) | (byte[3] << 24);

    /* Mark all patterns as dirty */
    for(i = 0; i < 512; ++i)    {
        smsvdp.pattern[i].dirty = 1;
        /* There's no use in clearing the textures, since the will be
            overwritten before they're used, anyway */
    }

    if(sms_console != CONSOLE_GG)   {
        for(i = 0; i < 0x20; ++i)   {
            update_local_pal_sms(i);
        }
    }
    else    {
        for(i = 0; i < 0x40; ++i)   {
            update_local_pal_gg(i);
        }
    }
}
