/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.
 
    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "CrabZ80.h"
#include "CrabZ80_tables.h"
#include "CrabZ80_macros.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>

static CrabZ80_t *cpu = NULL;

#ifndef CRABZ80_NO_READMAP_FALLBACK

#define FETCH_BYTE(name)    \
    uint8 name = cpu->readmap[cpu->pc.w >> 8] ? \
    cpu->readmap[cpu->pc.w >> 8][cpu->pc.w & 0xFF] : \
    cpu->mread(cpu->pc.w); \
    ++cpu->pc.w;

#define FETCH_WORD(name)    \
    uint16 name = cpu->readmap[cpu->pc.w >> 8] ? \
        cpu->readmap[cpu->pc.w >> 8][cpu->pc.w & 0xFF] : \
        cpu->mread(cpu->pc.w) | \
        ((cpu->readmap[(cpu->pc.w + 1) >> 8] ? \
          cpu->readmap[(cpu->pc.w + 1) >> 8][(cpu->pc.w + 1) & 0xFF] : \
          cpu->mread(cpu->pc.w + 1)) << 8); \
    cpu->pc.w += 2;

#else

#define FETCH_BYTE(name)    \
    uint8 name = cpu->readmap[cpu->pc.w >> 8][cpu->pc.w & 0xFF]; \
    ++cpu->pc.w;

#define FETCH_WORD(name)    \
    uint16 name = cpu->readmap[cpu->pc.w >> 8][cpu->pc.w & 0xFF] | \
        (cpu->readmap[(cpu->pc.w + 1) >> 8][(cpu->pc.w + 1) & 0xFF] << 8); \
    cpu->pc.w += 2;

#endif


#ifndef CRABZ80_BIG_SWITCHES

typedef void (*opcodeptr)();

static const opcodeptr _op_tab[0x100];
static const opcodeptr _opCB_tab[0x100];
static const opcodeptr _opDFD_tab[0x100];
static const opcodeptr _opED_tab[0x100];
static const opcodeptr _opDFDCB_tab[0x100];

#define EXECUTE(prefix, op) {   \
    uint8 _tmp = (op); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    _op##prefix##_tab[_tmp](); \
}

#define EXECUTE_PREFIX(prefix)  {   \
    FETCH_BYTE(_tmp); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    _op##prefix##_tab[_tmp](); \
}

#ifndef CRABZ80_NO_READMAP_FALLBACK

#define EXECUTE_DOUBLE_PREFIX() {   \
    uint16 addr = cpu->pc.w + 1; \
    uint8 _tmp = cpu->readmap[addr >> 8] ? \
        cpu->readmap[addr >> 8][addr & 0xFF] : \
        cpu->mread(addr); \
    cpu->cycles += cyclesDFDCB_tab[_tmp]; \
    _opDFDCB_tab[_tmp](); \
    ++cpu->pc.w; \
}

#else

#define EXECUTE_DOUBLE_PREFIX() {   \
    uint16 addr = cpu->pc.w + 1; \
    uint8 _tmp = cpu->readmap[addr >> 8][addr & 0xFF]; \
    cpu->cycles += cyclesDFDCB_tab[_tmp]; \
    _opDFDCB_tab[_tmp](); \
    ++cpu->pc.w; \
}

#endif /* CRABZ80_NO_READMAP_FALLBACK */

#else

static inline void __CrabZ80_execute(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeCB(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeDFD(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeED(uint8 op, CrabZ80_t *cpu);
static inline void __CrabZ80_executeDFDCB(uint8 op, CrabZ80_t *cpu);

#define EXECUTE(prefix, op) {   \
    uint8 _tmp = (op); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    __CrabZ80_execute ##prefix (_tmp, cpu); \
}

#define EXECUTE_PREFIX(prefix)  {   \
    FETCH_BYTE(_tmp); \
    cpu->cycles += cycles##prefix##_tab[_tmp]; \
    __CrabZ80_execute ##prefix (_tmp, cpu); \
}

#ifndef CRABZ80_NO_READMAP_FALLBACK

#define EXECUTE_DOUBLE_PREFIX() {   \
    uint16 addr = cpu->pc.w + 1; \
        uint8 _tmp = cpu->readmap[addr >> 8] ? \
        cpu->readmap[addr >> 8][addr & 0xFF] : \
        cpu->mread(addr); \
    cpu->cycles += cyclesDFDCB_tab[_tmp]; \
    __CrabZ80_executeDFDCB(_tmp, cpu); \
    ++cpu->pc.w; \
}

#else

#define EXECUTE_DOUBLE_PREFIX() {   \
    uint16 addr = cpu->pc.w + 1; \
    uint8 _tmp = cpu->readmap[addr >> 8][addr & 0xFF]; \
    cpu->cycles += cyclesDFDCB_tab[_tmp]; \
    __CrabZ80_executeDFDCB(_tmp, cpu); \
    ++cpu->pc.w; \
}

#endif /* CRABZ80_NO_READMAP_FALLBACK */

#endif

#include "CrabZ80_opcodes.h"

static uint8 CrabZ80_dummy_read(uint16 addr __attribute__((unused)))    {
    return 0;
}

static void CrabZ80_dummy_write(uint16 addr __attribute__((unused)), 
                                uint8 data __attribute__((unused))) {
}

#ifdef CRABZ80_USE_16_BIT_MEMORY
static uint16 CrabZ80_dummy_read16(uint16 addr __attribute__((unused))) {
    return 0;
}

static void CrabZ80_dummy_write16(uint16 addr __attribute__((unused)), 
                                  uint16 data __attribute__((unused)))  {
}
#endif

void CrabZ80_set_portread(CrabZ80_t *cpuz80, uint8 (*pread)(uint16 port))   {
    if(pread == NULL)
        cpuz80->pread = CrabZ80_dummy_read;
    else
        cpuz80->pread = pread;
}

void CrabZ80_set_memread(CrabZ80_t *cpuz80, uint8 (*mread)(uint16 addr))    {
    if(mread == NULL)
        cpuz80->mread = CrabZ80_dummy_read;
    else
        cpuz80->mread = mread;
}

void CrabZ80_set_portwrite(CrabZ80_t *cpuz80,
                           void (*pwrite)(uint16 port, uint8 data)) {
    if(pwrite == NULL)
        cpuz80->pwrite = CrabZ80_dummy_write;
    else
        cpuz80->pwrite = pwrite;
}

void CrabZ80_set_memwrite(CrabZ80_t *cpuz80,
                          void (*mwrite)(uint16 addr, uint8 data))  {
    if(mwrite == NULL)
        cpuz80->mwrite = CrabZ80_dummy_write;
    else
        cpuz80->mwrite = mwrite;
}

#ifdef CRABZ80_USE_16_BIT_MEMORY
void CrabZ80_set_memread16(CrabZ80_t *cpuz80,
                           uint16 (*mread16)(uint16 addr))  {
    if(mread16 == NULL)
        cpuz80->mread16 = CrabZ80_dummy_read16;
    else
        cpuz80->mread16 = mread16;
}

void CrabZ80_set_memwrite16(CrabZ80_t *cpuz80,
                            void (*mwrite16)(uint16 addr, uint16 data)) {
    if(mwrite16 == NULL)
        cpuz80->mwrite16 = CrabZ80_dummy_write16;
    else
        cpuz80->mwrite16 = mwrite16;
}
#endif

void CrabZ80_set_readmap(CrabZ80_t *cpuz80, uint8 *readmap[256])    {
    memcpy(cpuz80->readmap, readmap, 256 * sizeof(uint8 *));
}

void CrabZ80_init(CrabZ80_t *cpuz80)    {
    cpuz80->pread = CrabZ80_dummy_read;
    cpuz80->mread = CrabZ80_dummy_read;
    cpuz80->pwrite = CrabZ80_dummy_write;
    cpuz80->mwrite = CrabZ80_dummy_write;
}

void CrabZ80_reset(CrabZ80_t *cpuz80)   {
    cpuz80->pc.w = 0x0000;
    cpuz80->iff1 = 0;
    cpuz80->iff2 = 0;
    cpuz80->im = 0;
    cpuz80->halt = 0;
    cpuz80->r_top = 0;
    cpuz80->af.w = 0x0040;
    cpuz80->bc.w = 0x0000;
    cpuz80->de.w = 0x0000;
    cpuz80->hl.w = 0x0000;
    cpuz80->ix.w = 0xFFFF;
    cpuz80->iy.w = 0xFFFF;
    cpuz80->sp.w = 0xDFF0;
    cpuz80->ir.w = 0x0000;
    cpuz80->afp.w = 0x0000;
    cpuz80->bcp.w = 0x0000;
    cpuz80->dep.w = 0x0000;
    cpuz80->hlp.w = 0x0000;
    cpuz80->cycles = 0;
    cpuz80->cycles_in = 0;
    cpuz80->ei = 0;
    cpuz80->irq_pending = 0;
}

static void CrabZ80_take_nmi(CrabZ80_t *cpuz80) {
    if(cpuz80->halt)    {
        cpuz80->pc.w++;
        cpuz80->halt = 0;
    }

    ++cpuz80->ir.b.l;

    cpuz80->iff1 = 0;
    cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.h);
    cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.l);
    cpuz80->pc.w = 0x0066;
    cpuz80->cycles += 11;

    cpuz80->irq_pending &= 1;
}

static void CrabZ80_take_irq(CrabZ80_t *cpuz80) {
    if(cpuz80->halt)    {
        cpuz80->pc.w++;
        cpuz80->halt = 0;
    }

    ++cpuz80->ir.b.l;

    cpuz80->iff1 = cpuz80->iff2 = 0;

    switch(cpuz80->im)  {
        /* This case 0 is a hack specific to the Sega Master System 2/Game Gear
           technically, we should have some sort of function to give us the
           opcode to use in mode 0, but its not implemented yet. If someone
           really wants to use CrabZ80 for something that needs mode 0 or
           mode 2, let me know, and I'll implement them properly. Until then,
           or when I decide to do it, it'll stay this way. */
        case 0:
        case 1:
            cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.h);
            cpuz80->mwrite(--cpuz80->sp.w, cpuz80->pc.b.l);
            cpuz80->pc.w = 0x0038;
            cpuz80->cycles += 13;
            break;
        default:
            printf("Unimplemented Interrupt Mode: %d\n", cpuz80->im);
            assert(0);
    }
}

void CrabZ80_pulse_nmi(CrabZ80_t *cpuz80)   {
    cpuz80->irq_pending |= 2;
}

void CrabZ80_assert_irq(CrabZ80_t *cpuz80)  {
    cpuz80->irq_pending = 1;
}

void CrabZ80_clear_irq(CrabZ80_t *cpuz80)   {
    cpuz80->irq_pending &= 2;
}

void CrabZ80_release_cycles(CrabZ80_t *cpuz80)  {
    cpuz80->cycles_in = 0;
}

uint32 CrabZ80_execute(CrabZ80_t *cpuin, uint32 cycles) {
    uint32 real_cycles;
    CrabZ80_t *oldcpu;

    oldcpu = cpu;
    cpu = cpuin;

    cpu->cycles_in = cycles;

    while(cpu->cycles < cpu->cycles_in) {
        if(cpu->irq_pending & 2)    {
            CrabZ80_take_nmi(cpu);
        }
        else if(cpu->irq_pending && !cpu->ei && cpu->iff1)  {
            CrabZ80_take_irq(cpu);
        }

        cpu->ei = 0;
        {
            FETCH_BYTE(inst);
            EXECUTE(, inst);
            ++cpu->ir.b.l;
        }
    }

    real_cycles = cpu->cycles;
    cpu->cycles = 0;

    cpu = oldcpu;

    return real_cycles;
}
